# websites:
# https://pytorch.org/docs/stable/torchvision/transforms.html
# https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html#sphx-glr-beginner-blitz-cifar10-tutorial-py
# https://pytorch.org/hub/pytorch_vision_resnet/
# https://discuss.pytorch.org/t/normalize-each-input-image-in-a-batch-independently-and-inverse-normalize-the-output/23739
# https://pytorch.org/tutorials/beginner/transfer_learning_tutorial.html

import torch
import torchvision
import PIL
import math
import numpy as np
import os
import copy
from . import array as cc_array
from . import txt as cc_txt
from . import clock as cc_clock
from . import directory as cc_directory
from . import maths as cc_maths
from . import combinations as cc_combinations
from .strings import format_float_to_str as cc_strings_format_float_to_str


class Time:
    # TODO make range for each directory level
    def __init__(self, range_time, level, range_shifts=None, n_shifts=None):

        if not isinstance(range_time, cc_array.IntRange):
            self.range_time = cc_array.IntRange(range_time)
        else:
            self.range_time = range_time

        self.level = level

        if range_shifts is None:
            self.range_shifts = range_shifts
            self.random_shifts = False
        else:
            self.range_shifts = cc_array.IntRange(range_shifts)
            if self.range_shifts.len < 1:
                self.random_shifts = False
                self.range_shifts = None
            else:
                self.random_shifts = True

        self.possible_shifts = None
        self.n_possible_shifts = None
        self.n_shifts = None
        self.shifts = None

        self.use_tmp_shifts = False
        self.n_tmp_shifts = None
        self.tmp_shifts = None

        if self.random_shifts:
            self.possible_shifts = self.range_shifts.to_list()
            self.n_possible_shifts = self.possible_shifts.__len__()
            if n_shifts is not None:
                self.n_shifts = n_shifts
                self.n_tmp_shifts = self.n_possible_shifts * math.ceil(self.n_shifts / self.n_possible_shifts)
                if self.n_tmp_shifts > self.n_shifts:
                    self.use_tmp_shifts = True
                    self.tmp_shifts = self.possible_shifts * math.ceil(self.n_shifts / self.n_possible_shifts)
                else:
                    self.shifts = self.possible_shifts * math.ceil(self.n_shifts / self.n_possible_shifts)

    def set_n_shifts(self, n_shifts):

        if self.random_shifts and (n_shifts is not None):
            self.n_shifts = n_shifts
            self.n_tmp_shifts = self.n_possible_shifts * math.ceil(self.n_shifts / self.n_possible_shifts)
            if self.n_tmp_shifts > self.n_shifts:
                self.use_tmp_shifts = True
                self.tmp_shifts = self.possible_shifts * math.ceil(self.n_shifts / self.n_possible_shifts)
                self.shifts = None
            else:
                self.shifts = self.possible_shifts * math.ceil(self.n_shifts / self.n_possible_shifts)

    def next_shifts(self):

        if self.random_shifts and (self.n_shifts is not None):
            if self.use_tmp_shifts:
                self.shifts = np.random.permutation(self.tmp_shifts)[:self.n_shifts]
            else:
                self.shifts = np.random.permutation(self.shifts)


class Loader:
    """
    A class used to load batches of samples

    Extended description

    Attributes
    ----------
    directory_root : str
        A formatted string to print out what the animal says
    time : Time
        The name of the animal
    L, n_levels_directories : int
        The number of the directory tree in the dataset.
    conditions_directories : sequence of sequences of ints
        The l_th element of the sequence contains the conditions of the l_th directory level that will be loaded

    Methods
    -------
    load_batches_e(order_outputs=None)
        It is a batch generator. It load a batch of samples at the time.
    """

    # TODO add inter-sample levels and intra-sample levels
    def __init__(
            self, directory_root, level_classes, n_levels_directories=None,
            conditions_directories=None, batch_size=None, n_batches=None,
            shuffle=False, time=None, transforms=None, device=None,
            order_outputs='il'):
        """
        Parameters
        ----------
        directory_root : str
            The directory of the dataset
        level_classes : int
            The directory level of the classes in the directory tree of the dataset.
        n_levels_directories : int, optional
            number of levels of the directory tree of the dataset (default is None). It is required if
            conditions_directories is None.
        conditions_directories : None or sequence of Nones and sequences of ints, optional
            Conditions_directories can be a sequence or None (default is None). If it is a sequence, the l_th element
            of it is None or a sequence of ints. If the l_th element is a sequence of ints, it contains the conditions
            of the tree l_th level that will be loaded. If the l_th element is None all conditions of the l_th level
            will be loaded. If conditions_directories is None all conditions in all level will be loaded. It requires
            to be a sequence if n_levels_directories is None.
        order_outputs : str or sequence of str, optional
            The desired outputs of the self.load_batches_e(). Accepted values are "i", "l", "c","r", "a" or any
            combination of them like "ilcr", "ilr" (default is "il"). "i" stands for input samples, "l" for labels of
            the samples, "c" for combinations of the level conditions of the samples, "r" for the relative directories
            of the samples, "a" for the absolute directories of the samples.
        """

        self.directory_root = directory_root
        self.level_classes = level_classes

        self.time = time

        if n_levels_directories is None:
            self.n_levels_directories = len(conditions_directories)
        elif conditions_directories is None:
            self.n_levels_directories = n_levels_directories
            conditions_directories = [None] * self.n_levels_directories  # type: list

        self.L = self.n_levels_directories

        self.shuffle = shuffle

        if transforms is None:
            self.transforms = torchvision.transforms.Compose([torchvision.transforms.ToTensor()])
        else:
            self.transforms = transforms

        self.device = device

        self.conditions_directories_names = [None] * self.L  # type: list
        self.n_conditions_directories = [None] * self.L  # type: list
        self.conditions_directories = [None] * self.L  # type: list

        self.order_accepted_values = 'ilcra'
        if order_outputs is None:
            self.order_outputs = 'il'
            self.n_outputs = 2
        else:
            self.order_outputs = order_outputs
            self.n_outputs = len(self.order_outputs)
            for o in range(self.n_outputs):
                if not (self.order_outputs[o] in self.order_accepted_values):
                    raise ValueError('order_outputs')


        self.outputs = [None] * self.n_outputs  # type: list

        self.return_inputs_eb = 'i' in self.order_outputs
        self.return_labels_eb = 'l' in self.order_outputs
        self.return_combinations_eb = 'c' in self.order_outputs
        self.return_relative_directories_eb = 'r' in self.order_outputs
        self.return_absolute_directories_eb = 'a' in self.order_outputs

        directory_root_l = self.directory_root
        for l in range(self.L):
            self.conditions_directories_names[l] = os.listdir(directory_root_l)

            if conditions_directories[l] is None:
                self.n_conditions_directories[l] = len(self.conditions_directories_names[l])
                self.conditions_directories[l] = np.arange(self.n_conditions_directories[l])
            else:
                self.conditions_directories[l] = conditions_directories[l]
                self.n_conditions_directories[l] = len(self.conditions_directories[l])

            directory_root_l = os.path.join(directory_root_l, self.conditions_directories_names[l][0])

        if self.time is not None:
            if self.time.range_time.stop is None:
                self.time.range_time.stop = self.n_conditions_directories[self.time.level]
            self.conditions_directories[self.time.level] = (
                self.conditions_directories[self.time.level][self.time.range_time.to_slice()])
            self.n_conditions_directories[self.time.level] = len(self.conditions_directories[self.time.level])

        self.K = self.n_classes = self.n_conditions_directories[self.level_classes]
        self.V = self.n_variables = len(self.n_conditions_directories)
        self.n_samples = cc_maths.prod(self.n_conditions_directories)
        # self.n_samples = math.prod(self.n_conditions_directories)

        if batch_size is None:
            if n_batches is None:
                self.batch_size = self.n_samples
                self.n_batches = 1
            else:
                self.n_batches = n_batches
                self.batch_size = cc_maths.rint(self.n_samples / self.n_batches)
        else:
            self.batch_size = batch_size
            self.n_batches = cc_maths.rint(self.n_samples / self.batch_size)

        print('self.batch_size =', self.batch_size)
        print('self.n_samples =', self.n_samples)
        print('self.n_batches =', self.n_batches)
        print('self.time =', self.time)
        print('self.n_conditions_directories =', self.n_conditions_directories)
        # print('self.conditions_directories =', self.conditions_directories)

        if self.shuffle:
            self.batches_indexes = None
        else:
            self.batches_indexes = np.split(np.arange(self.n_samples), self.n_batches, axis=0)

        if (self.time is not None) and self.time.random_shifts:
            self.combinations_directories_no_shift = (
                cc_combinations.conditions_to_combinations(self.conditions_directories))
            self.combinations_directories = None
            # self.combinations_directories = np.copy(self.combinations_directories_no_shift)

            if self.time.n_shifts is None:
                self.time.set_n_shifts(self.n_samples)

            # self.time.next_shifts()
            # self.combinations_directories[:, self.time.level] += self.time.shifts
            if self.return_labels_eb:
                self.labels = torch.tensor(
                    self.combinations_directories_no_shift[slice(0, self.n_samples, 1), self.level_classes],
                    dtype=torch.int64, device=self.device)
            else:
                self.labels = None

        else:
            self.combinations_directories_no_shift = None
            self.combinations_directories = (
                cc_combinations.conditions_to_combinations(self.conditions_directories))
            if self.return_labels_eb:
                self.labels = torch.tensor(
                    self.combinations_directories[slice(0, self.n_samples, 1), self.level_classes],
                    dtype=torch.int64, device=self.device)
            else:
                self.labels = None

        if self.return_relative_directories_eb:
            self.relative_directories_eb = [None] * self.batch_size  # type: list
        else:
            self.relative_directories_eb = None

        if self.return_absolute_directories_eb:
            self.absolute_directories_eb = [None] * self.batch_size  # type: list
        else:
            self.absolute_directories_eb = None

        combination_directory_str_0 = [self.conditions_directories_names[l][0] for l in range(self.L)]
        directory_0 = os.path.join(self.directory_root, *combination_directory_str_0)
        image_0 = PIL.Image.open(directory_0)
        tensor_0 = self.transforms(image_0)
        # shape_sample_0 = np.asarray(tensor_0.shape)
        # self.shape_sample = shape_sample_0
        # self.shape_batch = np.empty(self.shape_sample.size + 1, self.shape_sample.dtype)
        # self.shape_batch[0] = self.batch_size
        # self.shape_batch[1:] = self.shape_sample
        # self.tensor_batch = torch.empty(tuple(self.shape_batch), dtype=tensor_0.dtype)
        shape_sample_0 = list(tensor_0.shape)
        self.shape_sample = shape_sample_0
        self.shape_batch = torch.Size(tuple([self.batch_size] + shape_sample_0))
        if self.return_inputs_eb:
            self.inputs_eb = torch.empty(self.shape_batch, dtype=tensor_0.dtype, device=self.device)
        else:
            self.inputs_eb = None

        self.n_dims_samples = self.shape_sample.__len__()
        self.n_dims_batch = self.n_dims_samples + 1
        self.indexes_batch = np.empty(self.n_dims_batch, dtype=object)
        self.indexes_batch[1:] = slice(0, None, 1)

        print('self.shape_batch =', self.shape_batch)
        print()

    def load_batches_e(self):
        """
        short description

        long description

        Returns
        -------
        out : generator
            A generator of lists of elements requested in order_outputs. The elements have the same order as in
            order_outputs. It generates a list per batch of input samples.

        Raises
        ------
        ValueError
            somenthing

        Examples
        --------
        example 1:

        >>> a = 2
        >>> b = 3
        >>> a + b
        5

        example 2:

        >>> a = 2
        >>> b = 3
        >>> a * b
        6

        """

        if (self.time is not None) and self.time.random_shifts:
            self.time.next_shifts()
            self.combinations_directories = np.copy(self.combinations_directories_no_shift)
            self.combinations_directories[:, self.time.level] += self.time.shifts

        if self.shuffle:
            self.batches_indexes = np.split(np.random.permutation(self.n_samples), self.n_batches, axis=0)

        for b in range(self.n_batches):
            if self.return_labels_eb:
                labels_eb = self.labels[self.batches_indexes[b]]

            combinations_eb = self.combinations_directories[self.batches_indexes[b], :]
            # batch_size_eb = batches_indexes_e.shape[0]
            for i in range(self.batch_size):
                combination_directory_ebi = combinations_eb[i, :]
                combination_directory_str_ebi = [
                    self.conditions_directories_names[l][combination_directory_ebi[l]] for l in range(self.L)]

                relative_directory_ebi = os.path.join(*combination_directory_str_ebi)
                absolute_directory_ebi = os.path.join(self.directory_root, relative_directory_ebi)

                self.indexes_batch[0] = i

                if self.return_inputs_eb:
                    inputs_ebi = PIL.Image.open(absolute_directory_ebi)
                    self.inputs_eb[tuple(self.indexes_batch)] = self.transforms(inputs_ebi)

                if self.return_relative_directories_eb:
                    self.relative_directories_eb[i] = relative_directory_ebi

                if self.return_absolute_directories_eb:
                    self.absolute_directories_eb[i] = absolute_directory_ebi

            for o in range(self.n_outputs):
                if self.order_outputs[o] == 'i':
                    self.outputs[o] = self.inputs_eb
                elif self.order_outputs[o] == 'l':
                    self.outputs[o] = labels_eb
                elif self.order_outputs[o] == 'c':
                    self.outputs[o] = combinations_eb
                elif self.order_outputs[o] == 'r':
                    self.outputs[o] = self.relative_directories_eb
                elif self.order_outputs[o] == 'a':
                    self.outputs[o] = self.absolute_directories_eb

            yield self.outputs


def train_classifier_with_early_stop(
        model, loader, criterion, optimizer, scheduler, I=20, directory_outputs=''):
    # timer = cc_clock.Timer()
    cc_timer = cc_clock.Timer()

    for key_loader_k in loader.keys():
        if key_loader_k == 'training' or key_loader_k == 'validation':
            pass
        else:
            raise ValueError('Unknown keys in loader')

    headers = [
        'Epoch', 'Unsuccessful Epochs', 'Training Loss', 'Training Accuracy',
        'Valuation Loss', 'Lowest Valuation Loss', 'Is Lower Loss',
        'Valuation Accuracy', 'Highest Valuation Accuracy', 'Is Higher Accuracy']

    n_columns = len(headers)
    new_line_stats = [None] * n_columns  # type: list

    stats = {
        'headers': {headers[k]: k for k in range(n_columns)},
        'n_columns': n_columns,
        'lines': []}

    directory_model = os.path.join(directory_outputs, 'model.pth')
    directory_model_state = os.path.join(directory_outputs, 'model_state.pth')
    directory_stats = os.path.join(directory_outputs, 'stats.csv')

    n_decimals_for_printing = 6

    best_model_wts = copy.deepcopy(model.state_dict())

    lowest_loss = math.inf
    lowest_loss_str = str(lowest_loss)
    highest_accuracy = -math.inf
    highest_accuracy_str = str(highest_accuracy)

    i = 0
    e = 0

    n_dashes = 110
    dashes = '-' * n_dashes
    print(dashes)

    while i < I:

        print('Epoch {e} ...'.format(e=e))

        stats['lines'].append(new_line_stats.copy())
        stats['lines'][e][stats['headers']['Epoch']] = e

        # Each epoch has a training and a validation phase
        # training phase
        model.train()  # Set model to training mode
        criterion.train()

        running_loss_e = 0.0
        running_corrects_e = 0

        b = 0
        # Iterate over data.
        for batch_eb, labels_eb in loader['training'].load_batches_e():
            # zero the parameter gradients
            optimizer.zero_grad()

            # forward
            # track history
            torch.set_grad_enabled(True)
            outputs = model(batch_eb)
            _, preds = torch.max(outputs, 1)
            loss_eb = criterion(outputs, labels_eb)

            # backward + optimize
            loss_eb.backward()
            optimizer.step()

            torch.set_grad_enabled(False)

            # # statistics
            # loss_float_eb = loss_eb.item()
            # # noinspection PyTypeChecker
            # corrects_eb = torch.sum(preds == labels_eb).item()
            # accuracy_eb = corrects_eb / batch_eb.shape[0]
            # loss_str_eb = cc_strings_format_float_to_str(loss_float_eb, n_decimals=n_decimals_for_printing)
            # accuracy_str_eb = cc_strings_format_float_to_str(accuracy_eb, n_decimals=n_decimals_for_printing)
            # print('Training. Epoch: {:d}. Batch {:d}. Loss: {:s}. Accuracy: {:s}.'.format(e, b, loss_str_eb, accuracy_str_eb))

            running_loss_e += loss_eb.item() * batch_eb.shape[0]
            # noinspection PyTypeChecker
            running_corrects_e += torch.sum(preds == labels_eb).item()

            b += 1

        # scheduler.step()

        loss_e = running_loss_e / loader['training'].n_samples
        accuracy_e = running_corrects_e / loader['training'].n_samples
        # loss_e = float(running_loss_e / loader['training'].n_samples)
        # accuracy_e = float(running_corrects_e / loader['training'].n_samples)

        loss_str_e = cc_strings_format_float_to_str(loss_e, n_decimals=n_decimals_for_printing)
        accuracy_str_e = cc_strings_format_float_to_str(accuracy_e, n_decimals=n_decimals_for_printing)

        print('Epoch: {:d}. Training.   Loss: {:s}. Accuracy: {:s}.'.format(e, loss_str_e, accuracy_str_e))

        stats['lines'][e][stats['headers']['Training Loss']] = loss_e
        stats['lines'][e][stats['headers']['Training Accuracy']] = accuracy_e
        # stats['Training Loss'].append(float(loss_e))
        # stats['Training Accuracy'].append(float(accuracy_e))

        # validation phase
        model.eval()  # Set model to evaluate mode

        criterion.eval()

        # zero the parameter gradients
        optimizer.zero_grad()

        torch.set_grad_enabled(False)

        running_loss_e = 0.0
        running_corrects_e = 0

        b = 0
        # Iterate over data.
        for batch_eb, labels_eb in loader['validation'].load_batches_e():
            # forward
            outputs = model(batch_eb)
            _, preds = torch.max(outputs, 1)
            loss_eb = criterion(outputs, labels_eb)

            # # statistics
            # loss_float_eb = loss_eb.item()
            # # noinspection PyTypeChecker
            # corrects_eb = torch.sum(preds == labels_eb).item()
            # accuracy_eb = corrects_eb / batch_eb.shape[0]
            # loss_str_eb = cc_strings_format_float_to_str(loss_float_eb, n_decimals=n_decimals_for_printing)
            # accuracy_str_eb = cc_strings_format_float_to_str(accuracy_eb, n_decimals=n_decimals_for_printing)
            # print('Validation. Epoch: {:d}. Batch {:d}. Loss: {:s}. Accuracy: {:s}.'.format(e, b, loss_str_eb, accuracy_str_eb))

            running_loss_e += loss_eb.item() * batch_eb.shape[0]
            # noinspection PyTypeChecker
            running_corrects_e += torch.sum(preds == labels_eb).item()

            b += 1

        loss_e = running_loss_e / loader['validation'].n_samples
        accuracy_e = running_corrects_e / loader['validation'].n_samples
        # loss_e = float(running_loss_e / loader['training'].n_samples)
        # accuracy_e = float(running_corrects_e / loader['training'].n_samples)

        loss_str_e = cc_strings_format_float_to_str(loss_e, n_decimals=n_decimals_for_printing)
        accuracy_str_e = cc_strings_format_float_to_str(accuracy_e, n_decimals=n_decimals_for_printing)

        stats['lines'][e][stats['headers']['Valuation Loss']] = loss_e
        stats['lines'][e][stats['headers']['Valuation Accuracy']] = accuracy_e

        if accuracy_e > highest_accuracy:
            highest_accuracy = accuracy_e
            highest_accuracy_str = cc_strings_format_float_to_str(highest_accuracy, n_decimals=n_decimals_for_printing)

            stats['lines'][e][stats['headers']['Is Higher Accuracy']] = 1
            stats['lines'][e][stats['headers']['Highest Valuation Accuracy']] = highest_accuracy
        else:
            stats['lines'][e][stats['headers']['Is Higher Accuracy']] = 0
            stats['lines'][e][stats['headers']['Highest Valuation Accuracy']] = highest_accuracy

        if loss_e < lowest_loss:

            lowest_loss = loss_e
            lowest_loss_str = cc_strings_format_float_to_str(lowest_loss, n_decimals=n_decimals_for_printing)
            i = 0
            stats['lines'][e][stats['headers']['Is Lower Loss']] = 1
            stats['lines'][e][stats['headers']['Unsuccessful Epochs']] = i
            stats['lines'][e][stats['headers']['Lowest Valuation Loss']] = lowest_loss

            best_model_wts = copy.deepcopy(model.state_dict())  # deep copy the model
            for directory_i in [directory_model, directory_model_state, directory_stats]:
                if os.path.isfile(directory_i):
                    os.remove(directory_i)

            torch.save(model, directory_model)
            torch.save(best_model_wts, directory_model_state)

            cc_txt.lines_to_csv_file(directory_stats, stats['lines'], stats['headers'])
        else:
            i += 1
            stats['lines'][e][stats['headers']['Is Lower Loss']] = 0
            stats['lines'][e][stats['headers']['Unsuccessful Epochs']] = i
            stats['lines'][e][stats['headers']['Lowest Valuation Loss']] = lowest_loss

            if os.path.isfile(directory_stats):
                os.remove(directory_stats)
            cc_txt.lines_to_csv_file(directory_stats, stats['lines'], stats['headers'])

        print('Epoch: {:d}. Validation. Loss: {:s}. Lowest Loss: {:s}. Accuracy: {:s}. Highest Accuracy: {:s}.'.format(
            e, loss_str_e, lowest_loss_str, accuracy_str_e, highest_accuracy_str))

        print('Epoch {e} - Unsuccessful Epochs {i}.'.format(e=e, i=i))

        e += 1
        print(dashes)

    print()
    E = e

    time_training = cc_timer.get_delta_time()

    print('Training completed in {d} days {h} hours {m} minutes {s} seconds'.format(
        d=time_training.days, h=time_training.hours,
        m=time_training.minutes, s=time_training.seconds))
    print('Number of Epochs: {E:d}'.format(E=E))
    print('Lowest Validation Loss: {:s}'.format(lowest_loss_str))
    print('Highest Validation Accuracy: {:s}'.format(highest_accuracy_str))

    # load best model weights
    model.load_state_dict(best_model_wts)
    return model, stats


def extract_features(model, loader, directory_dataset_features=''):
    cc_timer = cc_clock.Timer()

    n_decimals_for_printing = 6

    if model.training:
        model.eval()  # Set model to evaluate mode

    # Now set requires_grad to false
    for param in model.parameters():
        param.requires_grad = False

    torch.set_grad_enabled(False)

    # Iterate over data.
    for data_eb in loader.load_batches_e():

        samples_eb, relative_directories_eb = data_eb

        # forward
        outputs_eb = model(samples_eb)
        # outputs_eb = outputs_eb.numpy()

        # _, a = torch.max(outputs_eb, 1)
        # a = a.cpu().numpy()

        # 1. produce the directory_features
        relative_directories_features_eb = cc_directory.replace_extensions(relative_directories_eb, 'csv')
        absolute_directories_features_eb = cc_directory.conditions_to_directories(
            [[directory_dataset_features], relative_directories_features_eb], order_outputs='v')

        # 2. save features (make funtion array_to_csv_files(array, directories, axes='frc'))
        outputs_eb = torch.unsqueeze(outputs_eb, 1).cpu().numpy()
        # outputs_eb = torch.unsqueeze(outputs_eb, 1).numpy()
        cc_txt.array_to_csv_files(
            outputs_eb, 1, 2, [absolute_directories_features_eb], headers=None)

    time_extraction = cc_timer.get_delta_time()

    print('Test completed in {d} days {h} hours {m} minutes {s} seconds'.format(
        d=time_extraction.days, h=time_extraction.hours,
        m=time_extraction.minutes, s=time_extraction.seconds))


def test_classifier(model, loader, criterion, directory_test_output):
    cc_timer = cc_clock.Timer()

    headers_stats = ['N_samples', *['C_' + str(v) for v in range(loader.V)], 'Loss', 'Accuracy']

    n_columns_stats = len(headers_stats)
    line_stats = [loader.n_samples, *loader.n_conditions_directories, None, None]  # type: list

    stats = {
        'headers': {headers_stats[i]: i for i in range(n_columns_stats)},
        'lines': [line_stats]}

    headers_trials = [
        'ID_Trial',
        *['Condition_' + str(v) for v in range(loader.V)],
        'Label',
        *['Probability_' + str(k) for k in range(loader.K)],
        'Classification',
        'Correct_Classification'
    ]

    n_columns_trials = len(headers_trials)

    trials = {
        'headers': {headers_trials[i]: i for i in range(n_columns_trials)},
        'lines': None}

    n_decimals_for_printing = 6

    if model.training:
        model.eval()  # Set model to evaluate mode

    if criterion.training:
        criterion.eval()

    softmax = torch.nn.Softmax(dim=1)
    if softmax.training:
        softmax.eval()

    # Now set requires_grad to false
    for param_model in model.parameters():
        param_model.requires_grad = False

    for param_criterion in criterion.parameters():
        param_criterion.requires_grad = False

    for param_softmax in softmax.parameters():
        param_softmax.requires_grad = False

    torch.set_grad_enabled(False)

    running_loss_e = 0.0
    running_corrects_e = 0

    start_index_samples = 0
    stop_index_samples = 0

    index_combinations_e = np.empty(2, dtype=object)
    index_combinations_e[1] = slice(0, loader.V, 1)
    combinations_e = np.empty([loader.n_samples, loader.V], dtype=object)

    index_outputs_e = np.empty(2, dtype=object)
    index_outputs_e[1] = slice(0, loader.K, 1)
    outputs_e = np.empty([loader.n_samples, loader.K], dtype=object)

    index_labels_e = np.empty(2, dtype=object)
    index_labels_e[1] = 0
    labels_e = np.empty([loader.n_samples, 1], dtype=object)

    classifications_e = labels_e.copy()

    correct_classifications_e = labels_e.copy()

    id_trials = np.arange(loader.n_samples, dtype=object)[:, None]

    # b = 0
    # Iterate over data.
    for data_eb in loader.load_batches_e():
        samples_eb, labels_eb, combinations_eb = data_eb

        # forward
        outputs_eb = model(samples_eb)
        probabilities_eb = softmax(outputs_eb)
        _, classifications_eb = torch.max(outputs_eb, 1)
        correct_classifications_eb = (classifications_eb == labels_eb).long()
        loss_eb = criterion(outputs_eb, labels_eb)

        # todo: get probabilities
        # todo: fill trials['lines']

        stop_index_samples += samples_eb.shape[0]
        index_samples = slice(start_index_samples, stop_index_samples, 1)

        index_combinations_e[0] = index_samples
        combinations_e[tuple(index_combinations_e)] = combinations_eb.tolist()

        index_outputs_e[0] = index_samples
        outputs_e[tuple(index_outputs_e)] = probabilities_eb.tolist()

        index_labels_e[0] = index_samples
        labels_e[tuple(index_labels_e)] = labels_eb.tolist()

        classifications_e[tuple(index_labels_e)] = classifications_eb.tolist()

        correct_classifications_e[tuple(index_labels_e)] = correct_classifications_eb.tolist()

        start_index_samples = stop_index_samples

        running_loss_e += loss_eb.item() * samples_eb.shape[0]
        # noinspection PyTypeChecker
        running_corrects_e += torch.sum(correct_classifications_eb).item()

        # b += 1

    loss_e = running_loss_e / loader.n_samples
    accuracy_e = running_corrects_e / loader.n_samples

    stats['lines'][0][stats['headers']['Loss']] = loss_e
    stats['lines'][0][stats['headers']['Accuracy']] = accuracy_e

    trials['lines'] = np.concatenate(
        (id_trials, combinations_e, labels_e, outputs_e, classifications_e, correct_classifications_e),
        axis=1)

    loss_str_e = cc_strings_format_float_to_str(loss_e, n_decimals=n_decimals_for_printing)
    accuracy_str_e = cc_strings_format_float_to_str(accuracy_e, n_decimals=n_decimals_for_printing)

    print('Test. Loss: {:s}. Accuracy: {:s}.'.format(loss_str_e, accuracy_str_e))
    print()

    time_test = cc_timer.get_delta_time()

    print('Test completed in {d} days {h} hours {m} minutes {s} seconds'.format(
        d=time_test.days, h=time_test.hours,
        m=time_test.minutes, s=time_test.seconds))

    return stats, trials


class ResNetNoLastLayer(torchvision.models.resnet.ResNet):

    def __init__(self, name_resnet):

        if name_resnet == 'resnet18':
            super(ResNetNoLastLayer, self).__init__(torchvision.models.resnet.BasicBlock, [2, 2, 2, 2])
        elif name_resnet == 'resnet34':
            super(ResNetNoLastLayer, self).__init__(torchvision.models.resnet.BasicBlock, [3, 4, 6, 3])
        elif name_resnet == 'resnet50':
            super(ResNetNoLastLayer, self).__init__(torchvision.models.resnet.Bottleneck, [3, 4, 6, 3])
        elif name_resnet == 'resnet101':
            super(ResNetNoLastLayer, self).__init__(torchvision.models.resnet.Bottleneck, [3, 4, 23, 3])
        elif name_resnet == 'resnet152':
            super(ResNetNoLastLayer, self).__init__(torchvision.models.resnet.Bottleneck, [3, 8, 36, 3])

        elif name_resnet == 'resnext50_32x4d':
            kwargs = {'groups': 32, 'width_per_group': 4}
            super(ResNetNoLastLayer, self).__init__(torchvision.models.resnet.Bottleneck, [3, 4, 6, 3], **kwargs)

        elif name_resnet == 'resnext101_32x8d':
            kwargs = {'groups': 32, 'width_per_group': 8}
            super(ResNetNoLastLayer, self).__init__(torchvision.models.resnet.Bottleneck, [3, 4, 23, 3], **kwargs)

        elif name_resnet == 'wide_resnet50_2':
            kwargs = {'width_per_group': 64 * 2}
            super(ResNetNoLastLayer, self).__init__(torchvision.models.resnet.Bottleneck, [3, 4, 6, 3], **kwargs)

        elif name_resnet == 'wide_resnet101_2':
            kwargs = {'width_per_group': 64 * 2}
            super(ResNetNoLastLayer, self).__init__(torchvision.models.resnet.Bottleneck, [3, 4, 23, 3], **kwargs)

    def _forward_impl(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        return x


def load_resnet(name_resnet, last_layer=True, K=None, softmax=False, pretrained=False, device=None):

    if last_layer:
        if isinstance(pretrained, str):
            resnet = load_model(name_resnet, pretrained=False, device=None)
        else:
            resnet = load_model(name_resnet, pretrained=pretrained, device=None)

        if (K is None) or (K == resnet.fc.out_features):
            if softmax:
                resnet.fc = torch.nn.Sequential(resnet.fc, torch.nn.Softmax())
        else:
            num_ftrs = resnet.fc.in_features
            if softmax:
                resnet.fc = torch.nn.Sequential(torch.nn.Linear(num_ftrs, K), torch.nn.Softmax())
            else:
                # Here the size of each output sample is set to K.
                resnet.fc = torch.nn.Linear(num_ftrs, K)
        if isinstance(pretrained, str):
            state_dict = torch.load(pretrained)
            resnet.load_state_dict(state_dict)
    else:
        resnet = ResNetNoLastLayer(name_resnet)

        if isinstance(pretrained, str):
            if not ((K is None) or (K == resnet.fc.out_features)):
                num_ftrs = resnet.fc.in_features
                resnet.fc = torch.nn.Linear(num_ftrs, K)
            state_dict = torch.load(pretrained)
            resnet.load_state_dict(state_dict)
        elif pretrained:
            urls_resnets = {
                'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
                'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
                'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
                'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
                'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
                'resnext50_32x4d': 'https://download.pytorch.org/models/resnext50_32x4d-7cdf4587.pth',
                'resnext101_32x8d': 'https://download.pytorch.org/models/resnext101_32x8d-8ba56ff5.pth',
                'wide_resnet50_2': 'https://download.pytorch.org/models/wide_resnet50_2-95faca4d.pth',
                'wide_resnet101_2': 'https://download.pytorch.org/models/wide_resnet101_2-32ee1156.pth',
            }
            state_dict = torch.hub.load_state_dict_from_url(urls_resnets[name_resnet])
            resnet.load_state_dict(state_dict)
        delattr(resnet, 'fc')

    resnet = set_device(resnet, device)

    return resnet


def load_model(name_model, pretrained=False, device=None):
    # model = torch.hub.load('pytorch/vision:v0.6.0', 'resnet152', pretrained=True)

    # model = torchvision.models.resnet18()
    if isinstance(name_model, str):
        template_load_string = f"model = torchvision.models.{name_model:s}(pretrained=pretrained)"
    else:
        template_load_string = f"model = torchvision.models.{str(name_model):s}(pretrained=pretrained)"

    dict_globals = {'__builtins__': None}
    if isinstance(pretrained, str):
        dict_locals = {'torchvision': torchvision, 'pretrained': False}  # type: dict
    else:
        dict_locals = {'torchvision': torchvision, 'pretrained': pretrained}  # type: dict
    exec(template_load_string, dict_globals, dict_locals)
    model = dict_locals['model']

    if isinstance(pretrained, str):
        model.load_state_dict(torch.load(pretrained))

    model = set_device(model, device)

    return model


def set_device(tensor, device):

    # if isinstance(tensor, torch.Tensor):
    if isinstance(tensor, (torch.nn.Module, torch.Tensor)):
        if device is None:
            return tensor
        elif isinstance(device, (str, torch.device)):
            tensor = tensor.to(device)
            return tensor
        else:
            raise TypeError('device')
    else:
        raise TypeError('tensor')



import numpy as np
import sys
from ..maths import convert_to_int_or_float, rint
from ..array import FloatRange
from ..combinations import (
    conditions_to_combinations_on_the_fly as
    cc_conditions_to_combinations_on_the_fly)


def import_pygame():
    import pygame
    return pygame


pygame = import_pygame()


class Point:

    def __init__(self, point):

        if isinstance(point, Point):
            point_tmp = point.point
        elif isinstance(point, (np.ndarray, list, tuple)):
            point_tmp = point
        else:
            raise TypeError('point')

        # if isinstance(point_tmp, (np.ndarray, list, tuple)):
        x, y = point_tmp
        self.x = convert_to_int_or_float(x)
        self.y = convert_to_int_or_float(y)
        self.point = np.asarray([self.x, self.y])

    def convert_to_pygame_coordinates(self, size_frame):

        if isinstance(size_frame, (list, tuple)):
            half_size_frame = (np.asarray(size_frame) / 2)
        else:
            half_size_frame = (size_frame / 2)

        point_with_inverted_y = np.empty(2, dtype=self.point.dtype)
        point_with_inverted_y[0] = self.point[0]
        point_with_inverted_y[1] = -self.point[1]

        point_pg = PointPG((half_size_frame - .5) + point_with_inverted_y)

        return point_pg


class PointPG:

    def __init__(self, point, size_frame=None):

        if isinstance(point, PointPG):
            point_tmp = point.point
        elif isinstance(point, Point):
            point_tmp = point.convert_to_pygame_coordinates(size_frame).point
        elif isinstance(point, (np.ndarray, list, tuple)):
            point_tmp = point
        else:
            raise TypeError('point')

        # if isinstance(point_tmp, (np.ndarray, list, tuple)):
        x, y = point_tmp
        self.x = convert_to_int_or_float(x)
        self.y = convert_to_int_or_float(y)
        self.point = np.asarray([self.x, self.y])


class Line:

    def __init__(self, point_start, point_end):
        if isinstance(point_start, Point):
            self.point_start = point_start.point
        else:
            self.point_start = Point(point_start).point

        if isinstance(point_end, Point):
            self.point_end = point_end.point
        else:
            self.point_end = Point(point_end).point

    def convert_to_pygame_coordinates(self, size_frame):
        point_start_pg = Point(self.point_start).convert_to_pygame_coordinates(size_frame)
        point_end_pg = Point(self.point_end).convert_to_pygame_coordinates(size_frame)
        line_pg = LinePG(point_start_pg, point_end_pg)
        return line_pg


class LinePG:

    def __init__(self, point_start, point_end, size_frame=None):

        if isinstance(point_start, PointPG):
            self.point_start = point_start.point
        if isinstance(point_start, Point):
            self.point_start = point_start.convert_to_pygame_coordinates(size_frame).point
        else:
            self.point_start = PointPG(point_start).point

        if isinstance(point_end, PointPG):
            self.point_end = point_end.point
        if isinstance(point_end, Point):
            self.point_end = point_end.convert_to_pygame_coordinates(size_frame).point
        else:
            self.point_end = PointPG(point_end).point


class Rectangle:
    # (point, size, alignment_x=alignment_x, alignment_y=alignment_y)
    def __init__(self, point, size, alignment_x='c', alignment_y='c'):

        if isinstance(point, (list, tuple, np.ndarray)):
            point_tmp = Point(point)
        elif isinstance(point, Point):
            point_tmp = point
        elif isinstance(point, PointPG):
            raise TypeError('point')
        else:
            raise TypeError('point')

        self.width = self.w = convert_to_int_or_float(size[0])
        self.half_width = convert_to_int_or_float(self.w / 2)
        self.height = self.h = convert_to_int_or_float(size[1])
        self.half_height = convert_to_int_or_float(self.h / 2)

        alignment_x_low = alignment_x.lower()
        if alignment_x_low in ['c', 'center', 'centre']:
            self.centerx = point_tmp.x
            self.left = convert_to_int_or_float(self.centerx - (self.half_width - .5))
            self.right = convert_to_int_or_float(self.left + self.w)

        elif alignment_x_low in ['l', 'left']:
            self.left = point_tmp.x
            self.right = convert_to_int_or_float(self.left + self.w)
            self.centerx = convert_to_int_or_float(self.left + (self.half_width - .5))
        elif alignment_x_low in ['r', 'right']:
            self.right = point_tmp.x
            self.left = convert_to_int_or_float(self.right - self.w)
            self.centerx = convert_to_int_or_float(self.left + (self.half_width - .5))
        else:
            raise ValueError('alignment_x')

        alignment_y_low = alignment_y.lower()
        if alignment_y_low in ['c', 'center', 'centre']:
            self.centery = point_tmp.y
            self.bottom = convert_to_int_or_float(self.centery - (self.half_height - .5))
            self.top = convert_to_int_or_float(self.bottom + self.h)
        elif alignment_y_low in ['b', 'bottom']:
            self.bottom = point_tmp.y
            self.top = convert_to_int_or_float(self.bottom + self.h)
            self.centery = convert_to_int_or_float(self.bottom + (self.half_height - .5))
        elif alignment_y_low in ['t', 'top']:
            self.top = point_tmp.y
            self.bottom = convert_to_int_or_float(self.top - self.h)
            self.centery = convert_to_int_or_float(self.bottom + (self.half_height - .5))
        else:
            raise ValueError('alignment_y')

        self.bottomleft = np.asarray([self.left, self.bottom])
        self.bottomright = np.asarray([self.right, self.bottom])
        self.topleft = np.asarray([self.left, self.top])
        self.topright = np.asarray([self.right, self.top])
        self.center = np.asarray([self.centerx, self.centery])
        self.midleft = np.asarray([self.left, self.centery])
        self.midright = np.asarray([self.right, self.centery])
        self.midbottom = np.asarray([self.centerx, self.bottom])
        self.midtop = np.asarray([self.centerx, self.top])
        self.size = np.asarray([self.width, self.height])

    def convert_to_pygame_coordinates(self, size_frame):

        topleft_pg = Point(self.topleft).convert_to_pygame_coordinates(size_frame).point
        left_pg, top_pg = topleft_pg
        left_pg = rint(left_pg)
        top_pg = rint(top_pg)
        width = rint(self.width)
        height = rint(self.height)
        rect_pg = pygame.Rect(left_pg, top_pg, width, height)
        return rect_pg


class Cross:

    def __init__(self, center, size):

        if isinstance(center, Point):
            self.center = center.point
        else:
            self.center = Point(center).point

        if isinstance(size, (list, tuple)):
            len_size = len(size)
            if len_size == 1:
                self.size = np.asarray(size * 2)
            elif len_size == 2:
                self.size = np.asarray(size)
        elif isinstance(size, np.ndarray):
            len_shape = len(size.shape)
            if len_shape == 0:
                self.size = np.empty(2, dtype=size.dtype)
                self.size[:] = size
            elif len_shape == 1:
                if size.shape[0] == 1:
                    self.size = np.empty(2, dtype=size.dtype)
                    self.size[:] = size[0]
                elif size.shape[0] == 2:
                    self.size = size
        else:
            self.size = np.empty(2, dtype=type(size))
            self.size[:] = size

        self.half_size = self.size / 2

        # left, top = self.center - self.half_size
        # right, bottom = self.center + self.half_size
        left, top = self.center - (self.half_size - .5)
        right, bottom = self.center + (self.half_size - .5)

        self.left = convert_to_int_or_float(left)
        self.right = convert_to_int_or_float(right)
        self.top = convert_to_int_or_float(top)
        self.bottom = convert_to_int_or_float(bottom)

        self.point_start_horizontal_line = PointPG([self.left, self.center[1]])
        self.point_end_horizontal_line = PointPG([self.right, self.center[1]])

        self.point_start_vertical_line = Point([self.center[0], self.bottom])
        self.point_end_vertical_line = Point([self.center[0], self.top])

        self.horizontal_line = Line(self.point_start_horizontal_line, self.point_end_horizontal_line)
        self.vertical_line = Line(self.point_start_vertical_line, self.point_end_vertical_line)


class CrossPG:

    def __init__(self, center, size, size_frame):

        if isinstance(center, PointPG):
            self.center = center.point
        else:
            self.center = PointPG(center, size_frame=size_frame).point

        if isinstance(size, (list, tuple)):
            len_size = len(size)
            if len_size == 1:
                self.size = np.asarray(size * 2)
            elif len_size == 2:
                self.size = np.asarray(size)
        elif isinstance(size, np.ndarray):
            len_shape = len(size.shape)
            if len_shape == 0:
                self.size = np.empty(2, dtype=size.dtype)
                self.size[:] = size
            elif len_shape == 1:
                if size.shape[0] == 1:
                    self.size = np.empty(2, dtype=size.dtype)
                    self.size[:] = size[0]
                elif size.shape[0] == 2:
                    self.size = size
        else:
            self.size = np.empty(2, dtype=type(size))
            self.size[:] = size

        self.half_size = self.size / 2
        # left, top = self.center - self.half_size
        # right, bottom = self.center + self.half_size
        left, top = self.center - (self.half_size - .5)
        right, bottom = self.center + (self.half_size - .5)


        self.left = convert_to_int_or_float(left)
        self.right = convert_to_int_or_float(right)
        self.top = convert_to_int_or_float(top)
        self.bottom = convert_to_int_or_float(bottom)

        self.point_start_horizontal_line = PointPG([self.left, self.center[1]])
        self.point_end_horizontal_line = PointPG([self.right, self.center[1]])

        self.point_end_vertical_line = PointPG([self.center[0], self.top])
        self.point_start_vertical_line = PointPG([self.center[0], self.bottom])

        self.horizontal_line = LinePG(self.point_start_horizontal_line, self.point_end_horizontal_line)
        self.vertical_line = LinePG(self.point_start_vertical_line, self.point_end_vertical_line)


class Frame:

    def __init__(self, surface):
        self.surface = surface
        self.width, self.height = self.surface.get_size()
        self.w = self.width
        self.h = self.height
        self.half_width = self.half_w = convert_to_int_or_float(self.width / 2)
        self.half_height = self.half_h = convert_to_int_or_float(self.height / 2)
        self.size = np.asarray([self.w, self.h], dtype='i')
        self.half_size = np.asarray([self.half_w, self.half_h])

    def blit(self, source, point, alignment_x='c', alignment_y='c', area_sources=None, special_flags=0):

        # if isinstance(point, (list, tuple, np.ndarray)):
        #     point_pg = Point(point).convert_to_pygame_coordinates(self.size)
        # elif isinstance(point, Point):
        #     point_pg = point.convert_to_pygame_coordinates(self.size)
        # elif isinstance(point, PointPG):
        #     # point_pg = point
        #     raise TypeError('point')
        # else:
        #     raise TypeError('point')

        if isinstance(source, pygame.Surface):
            size_source = source.get_size()
        elif isinstance(source, Frame):
            size_source = source.size
        elif isinstance(source, StimText):
            size_source = source.frame.size
        else:
            raise TypeError('source')

        rect = Rectangle(point, size_source, alignment_x=alignment_x, alignment_y=alignment_y)
        rect_destination_pg = rect.convert_to_pygame_coordinates(self.size)

        # if isinstance(source, pygame.Surface):
        #     rect_destination_pg = source.get_rect()
        # elif isinstance(source, Frame):
        #     rect_destination_pg = source.surface.get_rect()
        # elif isinstance(source, StimText):
        #     rect_destination_pg = source.frame.surface.get_rect()
        # else:
        #     raise TypeError('source')


        # alignment_x_low = alignment_x.lower()
        # if alignment_x_low in ['c', 'center', 'centre']:
        #     rect_destination_pg.centerx = point_pg.x
        #     # rect_destination_pg.centerx = rint(point_pg.x + .5)
        # elif alignment_x_low in ['l', 'left']:
        #     rect_destination_pg.left = point_pg.x
        # elif alignment_x_low in ['r', 'right']:
        #     rect_destination_pg.right = point_pg.x
        # else:
        #     raise ValueError('alignment_x')
        #
        # alignment_y_low = alignment_y.lower()
        # if alignment_y_low in ['c', 'center', 'centre']:
        #     rect_destination_pg.centery = point_pg.y
        #     # rect_destination_pg.centery = rint(point_pg.y + .5)
        # elif alignment_y_low in ['b', 'bottom']:
        #     rect_destination_pg.bottom = point_pg.y
        # elif alignment_y_low in ['t', 'top']:
        #     rect_destination_pg.top = point_pg.y
        #     # rect_destination_pg.top = point_pg.y + 1
        # else:
        #     raise ValueError('alignment_y')

        if isinstance(source, pygame.Surface):
            self.surface.blit(
                source, rect_destination_pg,
                area=area_sources, special_flags=special_flags)
        elif isinstance(source, Frame):
            self.surface.blit(
                source.surface, rect_destination_pg,
                area=area_sources, special_flags=special_flags)
        elif isinstance(source, StimText):
            self.surface.blit(
                source.frame.surface, rect_destination_pg,
                area=area_sources, special_flags=special_flags)

    def blits(
            self, sources, point, alignment_x='c', alignment_y='c',
            area_sources=None, special_flags=0):

        n_sources = len(sources)
        for i in range(n_sources):
            self.blit(
                sources[i], point[i], alignment_x=alignment_x, alignment_y=alignment_y,
                area_sources=area_sources, special_flags=special_flags)

    def blits_in_different_columns_in_range_x(
            self, sources, range_x, point_y, alignment_x='c', alignment_y='c',
            area_sources=None, special_flags=0):

        n_sources = len(sources)
        myrange = FloatRange(range_x)
        alignment_x_low = alignment_x.lower()
        if alignment_x_low in ['c', 'center', 'centre']:
            points_x = myrange.centers_of_n_within_equal_ranges(n_sources)
        elif alignment_x_low in ['l', 'left']:
            points_x = myrange.starts_of_n_within_equal_ranges(n_sources)
        elif alignment_x_low in ['r', 'right']:
            points_x = myrange.stops_of_n_within_equal_ranges(n_sources)
        else:
            raise ValueError('alignment_x')

        for i in range(n_sources):
            self.blit(
                sources[i], [points_x[i], point_y], alignment_x=alignment_x, alignment_y=alignment_y,
                area_sources=area_sources, special_flags=special_flags)

    def blits_in_different_columns(
            self, sources, points_x, point_y, alignment_x='c', alignment_y='c',
            area_sources=None, special_flags=0):

        n_sources = len(sources)
        for i in range(n_sources):
            self.blit(
                sources[i], [points_x[i], point_y], alignment_x=alignment_x, alignment_y=alignment_y,
                area_sources=area_sources, special_flags=special_flags)

    def blits_in_different_rows(
            self, sources, point_x, points_y, alignment_x='c', alignment_y='c',
            area_sources=None, special_flags=0):

        n_sources = len(sources)
        for i in range(n_sources):
            self.blit(
                sources[i], [point_x, points_y[i]], alignment_x=alignment_x, alignment_y=alignment_y,
                area_sources=area_sources, special_flags=special_flags)

    def blits_in_different_columns_and_rows(
            self, sources, points_x, points_y, alignment_x='c', alignment_y='c',
            area_sources=None, special_flags=0):

        if not isinstance(points_x, (list, tuple, np.ndarray)):
            points_x = [points_x]
        if not isinstance(points_y, (list, tuple, np.ndarray)):
            points_y = [points_y]
        for (j, i), point_ji in cc_conditions_to_combinations_on_the_fly([points_x, points_y], order_outputs='iv'):
            self.blit(
                sources[j][i], point_ji, alignment_x=alignment_x, alignment_y=alignment_y,
                area_sources=area_sources, special_flags=special_flags)

    def fill(self, color, rect=None):
        if rect is None:
            rect_pg = None
        else:
            rect_pg = rect.convert_to_pygame_coordinates(self.size)

        self.surface.fill(color, rect_pg)
        return None

    def draw_line_from_points(self, point_start, point_end, color, width=1):

        if isinstance(point_start, Point):
            point_start_pg = point_start.convert_to_pygame_coordinates(self.size)
        elif isinstance(point_start, PointPG):
            point_start_pg = point_start
        else:
            point_start_pg = Point(point_start).convert_to_pygame_coordinates(self.size)

        if isinstance(point_end, Point):
            point_end_pg = point_end.convert_to_pygame_coordinates(self.size)
        elif isinstance(point_end, PointPG):
            point_end_pg = point_end
        else:
            point_end_pg = Point(point_end).convert_to_pygame_coordinates(self.size)

        line_pg = LinePG(point_start_pg, point_end_pg)
        pygame.draw.line(self.surface, color, line_pg.point_start, line_pg.point_end, width)

    def draw_line_from_line(self, line, color, width=1):
        if isinstance(line, Line):
            line_pg = line.convert_to_pygame_coordinates(self.size)
        elif isinstance(line, LinePG):
            line_pg = line
        else:
            raise TypeError('line')

        pygame.draw.line(self.surface, color, line_pg.point_start, line_pg.point_end, width)

    def draw_cross_from_point(self, point, size, color, width=1, alignment_x='c', alignment_y='c'):

        if isinstance(point, Point):
            point_pg = point.convert_to_pygame_coordinates(self.size)
        elif isinstance(point, PointPG):
            point_pg = point
        else:
            point_pg = Point(point).convert_to_pygame_coordinates(self.size)
        # TODO: cross alignment_x='c', alignment_y='c'
        cross_pg = CrossPG(point_pg, size, self.size)
        # draw horizontal line
        self.draw_line_from_line(cross_pg.horizontal_line, color, width=width)
        # draw horizontal line
        self.draw_line_from_line(cross_pg.vertical_line, color, width=width)

    def subsurface(self, rect):
        rect_pg = rect.convert_to_pygame_coordinates(self.size)
        my_subsurface = Frame(self.surface.subsurface(rect_pg))
        return my_subsurface

    def get_rect(self):
        left = -self.half_width
        bottom = -self.half_height
        rect = Rectangle([left, bottom], [self.width, self.height], alignment_x='l', alignment_y='b')
        return rect


def initiate_font():
    pygame.font.init()


class StimText:

    def __init__(
            self, text, size, color,
            alignment_rows='c', name=None, bold=False, italic=False, distance_rows=1,
            antialias=False, background=None):

        if isinstance(text, str):
            self.text = text
        elif isinstance(text, (list, tuple, np.ndarray)):
            self.text = '\n'.join(text)
        else:
            raise TypeError('text')

        self.rows = self.text.splitlines()
        self.n_rows = self.R = len(self.rows)

        self.size = size
        self.color = color
        self.alignment_rows = alignment_rows
        self.name = name
        self.bold = bold
        self.italic = italic
        self.distance_rows = distance_rows
        self.antialias = antialias
        self.background = background

        font = pygame.font.SysFont(name, size, bold=bold, italic=italic)
        if self.R == 1:
            self.frame = Frame(font.render(self.rows[0], self.antialias, self.color, self.background))
        else:
            frames: list = [None] * self.R
            widths_lines = np.empty(self.R, dtype='i')
            heights_lines = np.empty(self.R, dtype='i')
            for r in range(self.R):
                frames[r] = Frame(font.render(self.rows[r], self.antialias, self.color, self.background))
                widths_lines[r], heights_lines[r] = frames[r].size

            max_width_lines = int(widths_lines.max(initial=0))
            max_height_lines = int(heights_lines.max(initial=0))
            height = ((max_height_lines * self.distance_rows) * (self.R - 1)) + max_height_lines
            self.frame = Frame(pygame.Surface([max_width_lines, height]))

            if self.background is None:
                diff_color = 50
                threshould = 255 - diff_color
                background_tmp = [0, 0, 0]
                for c in range(3):
                    if self.color[c] <= threshould:
                        background_tmp[c] = self.color[c] + diff_color
                    elif self.color[c] > threshould:
                        background_tmp[c] = self.color[c] - diff_color

                # background_tmp = [self.color[c] for c in range(3)]
                # if 0 <= background_tmp[0] < 255:
                #     background_tmp[0] += 1
                # elif background_tmp[0] == 255:
                #     background_tmp[0] -= 1
                # else:
                #     raise ValueError('color[0]')
                # background_tmp = [50, 200, 50]
                # if 255 in self.color:
                #     background_tmp = [0, 0, 0]
                # else:
                #     background_tmp = [255, 255, 255]

                self.frame.fill(background_tmp)
                # self.frame.fill(self.color)
                self.frame.surface.set_colorkey(background_tmp)
            else:
                self.frame.surface.fill(self.background)

            left_first_line = -self.frame.half_width
            bottom_first_line = self.frame.half_height - max_height_lines
            rect_first_line = Rectangle(
                [left_first_line, bottom_first_line],
                [self.frame.width, max_height_lines],
                alignment_x='l', alignment_y='b')
            surface_first_line = self.frame.subsurface(rect_first_line)

            bottom_other_lines = -self.frame.half_height
            height_other_lines = self.frame.height - max_height_lines
            rect_other_lines = Rectangle(
                [left_first_line, bottom_other_lines],
                [self.frame.width, height_other_lines],
                alignment_x='l', alignment_y='b')
            surface_other_lines = self.frame.subsurface(rect_other_lines)

            alignment_rows_lower = alignment_rows.lower()
            if alignment_rows_lower in ['c', 'center', 'centre']:
                x_lines = 0
            elif alignment_rows_lower in ['l', 'left']:
                x_lines = -max_width_lines / 2
            elif alignment_rows_lower in ['r', 'right']:
                x_lines = max_width_lines / 2
            else:
                raise ValueError('alignment_x')

            surface_first_line.blit(
                frames[0], [x_lines, surface_first_line.half_height], alignment_x=alignment_rows, alignment_y='t')

            num_sources = np.arange(self.R - 1)
            y_lines = ((num_sources[::-1] - num_sources.mean() - 0.5) * (max_height_lines * self.distance_rows))

            surface_other_lines.blits_in_different_rows(
                frames[1:], x_lines, y_lines, alignment_x=alignment_rows, alignment_y='b')

    # def draw(self, destination, point, area_sources=None, special_flags=0):
    #
    #     if isinstance(point, PointPG):
    #         raise TypeError('point')
    #
    #     num_sources = np.arange(self.R)
    #     points_y = point[1] + ((num_sources[::-1] - num_sources.mean()) * self.size)
    #     point_x = point[0]
    #
    #     if isinstance(destination, pygame.Surface):
    #         Frame(destination).blits_in_different_rows(
    #             self.frames, point_x, points_y,
    #             area_sources=area_sources, special_flags=special_flags)
    #     elif isinstance(destination, Frame):
    #         destination.blits_in_different_rows(
    #             self.frames, point_x, points_y,
    #             area_sources=area_sources, special_flags=special_flags)


def list_of_text_to_list_of_stimtext(
        text, size, same_color=True, color=None,
        alignment_rows='c', name=None, bold=False, italic=False, distance_rows=1,
        antialias=False, background=None):

    if color is None:
        color = [255, 255, 255]
        same_color = True

    n_texts = len(text)
    frames: list = [None] * n_texts

    if same_color:
        for i in range(n_texts):
            frames[i] = StimText(
                text[i], size, color, alignment_rows=alignment_rows,
                name=name, bold=bold, italic=italic,
                distance_rows=distance_rows, antialias=antialias, background=background)
    else:
        for i in range(n_texts):
            frames[i] = StimText(
                text[i], size, color[i], alignment_rows=alignment_rows,
                name=name, bold=bold, italic=italic,
                distance_rows=distance_rows, antialias=antialias, background=background)

    return frames


def surfaces_to_sizes(surfaces):
    n_surfaces = len(surfaces)
    sizes: list = [None] * n_surfaces
    for i in range(n_surfaces):
        sizes[i] = surfaces[i].get_size()
    return sizes


def print_names_fonts():
    names = pygame.font.get_fonts()
    print('There are {n} fonts:'.format(n=len(names)))
    for name_i in names:
        print(name_i)


def font_names():
    names = pygame.font.get_fonts()
    return names


def update(rect=None):
    if rect is None:
        pygame.display.flip()
    else:
        info = pygame.display.Info()
        size_screen = [info.current_w, info.current_h]
        rect_pg = rect.convert_to_pygame_coordinates(size_screen)
        pygame.display.update(rect_pg)
    return None


def flip_until_key_is_pressed(key):
    wait = True
    while wait:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()

            elif event.type == pygame.KEYDOWN:

                if event.key == pygame.K_ESCAPE:
                    pygame.quit()
                    sys.exit()
                elif event.key == key:
                    wait = False

        pygame.display.flip()



import os
import sys
from . import combinations as cc_combinations
from . import maths as cc_maths


def n_directories_up(directory, n=1):
    
    while n > 0:
        directory = os.path.dirname(directory)
        n -= 1

    return directory


def get_root_directory_of_running_script():
    directory_scripts = os.path.realpath(sys.argv[0])
    root_directory = os.path.dirname(directory_scripts)
    return root_directory


def conditions_to_directories_on_the_fly(conditions_directories, order_outputs='v'):
    """
    Parameters
    ----------
    order_outputs : str or sequence of str, optional
        The desired outputs. Accepted values are "v", "i" or any combination of them like "vi", "iv" (default is "vi").
        "v" stands for combination_values_i and "i" for combination_indexes_i.
    """
    order_accepted_values = 'vi'
    if order_outputs is None:
        order_outputs = 'v'
        n_outputs = 2
    else:
        n_outputs = len(order_outputs)
        if n_outputs < 1:
            raise ValueError('order_outputs')
        else:
            for o in range(n_outputs):
                if not (order_outputs[o] in order_accepted_values):
                    raise ValueError('order_outputs')
    if n_outputs > 1:
        outputs_i = [None] * n_outputs  # type: list

    for combinations_i in cc_combinations.conditions_to_combinations_on_the_fly(
            conditions_directories, dtype='U', order_outputs=order_outputs):
        if n_outputs > 1:
            for o in range(n_outputs):
                if order_outputs[o] == 'v':
                    combination_directories_i = combinations_i[o]
                    directory_i = os.path.join(*combination_directories_i)
                    outputs_i[o] = directory_i
                elif order_outputs[o] == 'i':
                    combination_indexes_i = combinations_i[o]
                    outputs_i[o] = combination_indexes_i
            yield outputs_i
        elif n_outputs == 1:
            if order_outputs[0] == 'v':
                combination_directories_i = combinations_i
                directory_i = os.path.join(*combination_directories_i)
                yield directory_i
            elif order_outputs[0] == 'i':
                combination_indexes_i = combinations_i
                yield combination_indexes_i


def conditions_to_directories(conditions_directories, order_outputs='v'):

    n_conditions = cc_combinations.conditions_to_n_conditions(conditions_directories)
    n_directories = cc_maths.prod(n_conditions)
    # n_variables = len(n_conditions)

    order_accepted_values = 'vi'
    if order_outputs is None:
        order_outputs = 'v'
        n_outputs = 2
    else:
        n_outputs = len(order_outputs)
        if n_outputs < 1:
            raise ValueError('order_outputs')
        else:
            for o in range(n_outputs):
                if not (order_outputs[o] in order_accepted_values):
                    raise ValueError('order_outputs')
    if n_outputs > 1:
        outputs_i = [[None] * n_directories] * n_outputs  # type: list
    else:
        outputs_i = [None] * n_directories  # type: list

    # directories = [None] * n_directories  # type: list
    i = -1
    for combinations_i in conditions_to_directories_on_the_fly(conditions_directories, order_outputs=order_outputs):
        i += 1
        # directories[i] = directory_i

        if n_outputs > 1:
            for o in range(n_outputs):
                if order_outputs[o] == 'v':
                    combination_directories_i = combinations_i[o]
                    outputs_i[o][i] = combination_directories_i
                elif order_outputs[o] == 'i':
                    combination_indexes_i = combinations_i[o]
                    outputs_i[o][i] = combination_indexes_i
        elif n_outputs == 1:
            if order_outputs[0] == 'v':
                combination_directories_i = combinations_i
                outputs_i[i] = combination_directories_i
            elif order_outputs[0] == 'i':
                combination_indexes_i = combinations_i
                outputs_i[i] = combination_indexes_i

    return outputs_i


def remove_extension(directory):
    dirname = os.path.dirname(directory)
    basename_with_extension = os.path.basename(directory)
    basename_no_extension = os.path.splitext(basename_with_extension)[0]
    directory_no_extension = os.path.join(dirname, basename_no_extension)
    return directory_no_extension


def add_extension(directory, extension):
    if extension[0] == '.':
        directory_with_extension = ''.join([directory, extension])
    else:
        directory_with_extension = '.'.join([directory, extension])
    return directory_with_extension


def replace_extension(directory, extension):
    directory_no_extension = remove_extension(directory)
    directory_with_extension = add_extension(directory_no_extension, extension)
    return directory_with_extension


def replace_extensions(directories, extension):
    n_directories = len(directories)
    for d in range(n_directories):
        directories[d] = replace_extension(directories[d], extension)
    return directories



import cv2


def read_image(directory, order='RGB'):
    array_bgr = cv2.imread(directory, cv2.IMREAD_COLOR)
    order_l = order.upper()
    if order_l == 'RGB':
        array_rgb = cv2.cvtColor(array_bgr, cv2.COLOR_BGR2RGB)
        # array_rgb = array_bgr[:, :, slice(-1, -4, -1)]
        return array_rgb
    elif order_l == 'BGR':
        return array_bgr
    else:
        raise ValueError('order')



import numpy as np
import math
import itertools
from . import format as cc_format
from . import array as cc_array
from . import maths as cc_maths


def n_conditions_to_combinations(
        n_conditions,
        axis_combinations=0,
        n_repetitions_combinations=1,
        order_variables='rl',
        variables_in_order=None,
        dtype='i'):

    # order_variables is the order of the variables by which they change their own conditions.
    # order_variables can either be 'rl' for right-to-left, 'lr' for left-to-right or 'c' for custom.
    # If order_variables=='c', then variables_in_order must be a list, a tuple or an 1-d array containing the variables
    # in the custom order by which the variables change their own conditions:
    #
    # Requirements:
    # 1) order_variables == 'rl' or order_variables == 'lr' or order_variables == 'c';
    # If order_variables=='c', then:
    #     2) len(variables_in_order) == len(n_conditions);
    #     3) variables_in_order must contain all integers from 0 to (len(n_conditions)-1);
    #     4) variables_in_order cannot contain repeated integers.

    n_variables = len(n_conditions)
    if n_variables > 0:
        variables = np.arange(n_variables)
        # n_conditions = np.asarray(n_conditions, dtype=int)
        # conditions = n_conditions_to_conditions(n_conditions)
        n_combinations = cc_maths.prod(n_conditions) * n_repetitions_combinations
        if n_combinations < 0:
            n_conditions = n_conditions.astype(np.int64)
            n_combinations = cc_maths.prod(n_conditions) * n_repetitions_combinations

        # axis_combinations = int(not(bool(axis_variables)))
        axis_variables = int(not(bool(axis_combinations)))
        shape_combinations = np.empty(2, dtype='i')
        shape_combinations[axis_combinations] = n_combinations
        shape_combinations[axis_variables] = n_variables

        combinations = np.empty(shape_combinations, dtype=dtype)

        if order_variables == 'rl':
            variables_in_order = variables[::-1]
        elif order_variables == 'lr':
            variables_in_order = variables
        elif order_variables == 'c':
            variables_in_order = np.asarray(variables_in_order, dtype='i')
            # check requirement 2
            if len(variables_in_order) != n_variables:
                raise ValueError('The following condition must be met:\nlen(variables_in_order) == len(n_conditions).')
            # check requirement 3
            for v in range(n_variables):
                if v not in variables_in_order:
                    raise ValueError('variables_in_order must contain all integers from 0 to (len(n_conditions)-1).\n'
                                     '{} is not in variables_in_order.'.format(v))
            # check requirement 4
            for v in variables_in_order:
                if np.sum(v == variables_in_order) > 1:
                    raise ValueError('variables_in_order cannot contain repeated integers.\n'
                                     '{} is a repeated integer.'.format(v))
        else:
            # check requirement 1
            raise ValueError(
                'order_variables can either be \'rl\' for right-to-left, '
                '\'lr\' for left-to-right or \'c\' for custom.\n'
                'Now, order_variables = {}'.format(order_variables))

        i_variable = variables_in_order[0]
        indexes_combinations = np.empty(2, dtype=object)
        indexes_combinations[axis_variables] = i_variable
        for i_condition in range(n_conditions[i_variable]):
            indexes_combinations[axis_combinations] = slice(i_condition, n_combinations, n_conditions[i_variable])
            combinations[tuple(indexes_combinations)] = i_condition

        indexes_combinations[axis_combinations] = slice(None)

        # combinations[tuple(indexes_combinations)] = np.expand_dims(
        #     cc_array.pad_array_from_n_samples_target(conditions[-1], n_samples_target=n_combinations),
        #     axis=axis_variables)
        cumulative_n_combinations = n_conditions[i_variable]
        for i_variable in variables_in_order[1:]:
            cumulative_combinations = np.empty(
                cumulative_n_combinations * n_conditions[i_variable], combinations.dtype)
            for i_condition in range(n_conditions[i_variable]):
                cumulative_combinations[
                    slice(i_condition * cumulative_n_combinations,
                          (i_condition + 1) * cumulative_n_combinations)] = i_condition

            indexes_combinations[axis_variables] = i_variable
            if cumulative_combinations.size < n_combinations:
                combinations[tuple(indexes_combinations)] = cc_array.pad_array_from_n_samples_target(
                    cumulative_combinations, n_samples_target=n_combinations)
            elif cumulative_combinations.size == n_combinations:
                combinations[tuple(indexes_combinations)] = cumulative_combinations
            else:
                raise Exception('something wrong')
            cumulative_n_combinations *= n_conditions[i_variable]

        # if change_dtype:
        #     combinations = combinations.astype(dtype_end)
    else:
        combinations = []
    return combinations


def n_conditions_to_combinations_on_the_fly(
        n_conditions, order_variables='rl', variables_in_order=None, dtype='i'):
    try:
        n_variables = len(n_conditions)
        if not isinstance(n_conditions, np.ndarray):
            n_conditions = np.asarray(n_conditions, dtype='i')
    except TypeError:
        n_conditions = np.asarray([n_conditions], dtype='i')
        n_variables = 1

    # n_combinations = int(np.prod(n_conditions))
    n_combinations = cc_maths.prod(n_conditions)

    combination_i = np.empty(n_variables, dtype=dtype)
    combination_i[:] = 0

    yield combination_i

    for i in range(1, n_combinations):

        v = n_variables - 1
        increase_condition_v = True

        while increase_condition_v:
            combination_i[v] += 1
            if combination_i[v] >= n_conditions[v]:
                combination_i[v] = 0
                v -= 1
            else:
                increase_condition_v = False

        yield combination_i


def conditions_to_combinations_on_the_fly(
        conditions_raw, order_variables='rl', variables_in_order=None, dtype=None, order_outputs='vi'):
    """
    Parameters
    ----------
    order_outputs : str or sequence of str, optional
        The desired outputs. Accepted values are "v", "i" or any combination of them like "vi", "iv" (default is "vi").
        "v" stands for combination_values_i and "i" for combination_indexes_i.
    """

    if not isinstance(conditions_raw, np.ndarray):
        conditions = np.asarray(conditions_raw, dtype=object)
    else:
        conditions = conditions_raw

    n_conditions = conditions_to_n_conditions(conditions)
    n_combinations = cc_maths.prod(n_conditions)
    n_variables = len(n_conditions)

    combination_indexes_i = np.empty(n_variables, dtype='i')
    combination_indexes_i[:] = 0

    order_accepted_values = 'vi'
    if order_outputs is None:
        order_outputs = 'vi'
        n_outputs = 2
    else:
        n_outputs = len(order_outputs)
        if n_outputs < 1:
            raise ValueError('order_outputs')
        else:
            for o in range(n_outputs):
                if not (order_outputs[o] in order_accepted_values):
                    raise ValueError('order_outputs')

    outputs_i = [None] * n_outputs  # type: list

    for o in range(n_outputs):

        if order_outputs[o] == 'v':

            if dtype is None:
                combination_values_i = [None] * n_variables
                dtype_np = None
            else:
                dtype_np = np.dtype(dtype)
                if dtype_np.kind == 'U':
                    combination_values_i = np.empty(n_variables, dtype='O')
                else:
                    combination_values_i = np.empty(n_variables, dtype=dtype)

            for v in range(n_variables):
                combination_values_i[v] = conditions[v][combination_indexes_i[v]]

            if dtype_np is None:
                combination_values_i = np.asarray(combination_values_i)
                dtype_np = combination_values_i.dtype
                if dtype_np.kind == 'U':
                    combination_values_i = combination_values_i.astype('O')

            if dtype_np.kind == 'U':
                outputs_i[o] = combination_values_i.astype('U')
            else:
                outputs_i[o] = combination_values_i

        elif order_outputs[o] == 'i':

            outputs_i[o] = combination_indexes_i

    if n_outputs == 1:
        yield outputs_i[0]
    else:
        yield outputs_i

    for i in range(1, n_combinations):

        v = n_variables - 1
        increase_condition_v = True

        while increase_condition_v:
            combination_indexes_i[v] += 1
            if combination_indexes_i[v] >= n_conditions[v]:
                combination_indexes_i[v] = 0
                v -= 1
            else:
                increase_condition_v = False

        for o in range(n_outputs):

            if order_outputs[o] == 'v':

                for v in range(n_variables):
                    combination_values_i[v] = conditions[v][combination_indexes_i[v]]

                if dtype_np.kind == 'U':
                    outputs_i[o] = combination_values_i.astype('U')
                else:
                    outputs_i[o] = combination_values_i

            elif order_outputs[o] == 'i':

                outputs_i[o] = combination_indexes_i

        if n_outputs == 1:
            yield outputs_i[0]
        else:
            yield outputs_i


# def conditions_to_combinations_on_the_fly(
#         conditions, order_variables='rl', variables_in_order=None, dtype=None):
#
#     for _, combination_i in conditions_to_combinations_on_the_fly_with_indexes(
#             conditions, order_variables=order_variables,
#             variables_in_order=variables_in_order, dtype=dtype):
#
#         yield combination_i


def conditions_to_combinations(
        conditions,
        axis_combinations=0,
        n_repetitions_combinations=1,
        order_variables='rl',
        variables_in_order=None,
        dtype=None):

    # order_variables is the order of the variables by which they change their own conditions.
    # order_variables can either be 'rl' for right-to-left, 'lr' for left-to-right or 'c' for custom.
    # If order_variables=='c', then variables_in_order must be a list, a tuple or an 1-d array containing the variables
    # in the custom order by which the variables change their own conditions:
    #
    # Requirements:
    # 1) order_variables == 'rl' or order_variables == 'lr' or order_variables == 'c';
    # If order_variables=='c', then:
    #     2) len(variables_in_order) == len(conditions);
    #     3) variables_in_order contains all integers from 0 to (len(conditions)-1);
    #     4) variables_in_order cannot contain repeated integers.

    n_variables = len(conditions)
    if n_variables > 0:
        variables = np.arange(n_variables)
        # conditions = np.asarray(conditions, dtype=object)
        n_conditions = conditions_to_n_conditions(conditions)
        n_combinations = cc_maths.prod(n_conditions) * n_repetitions_combinations
        if n_combinations < 0:
            n_conditions = n_conditions.astype(np.int64)
            n_combinations = cc_maths.prod(n_conditions) * n_repetitions_combinations

        if dtype is None:
            dtype = type(conditions[0][0])

        change_dtype = False
        if dtype == str:
            dtype_end = str
            dtype = object
            change_dtype = True

        axis_variables = int(not(bool(axis_combinations)))
        shape_combinations = np.empty(2, dtype=int)
        shape_combinations[axis_combinations] = n_combinations
        shape_combinations[axis_variables] = n_variables

        combinations = np.empty(shape_combinations, dtype=dtype)

        if order_variables == 'rl':
            variables_in_order = variables[::-1]
        elif order_variables == 'lr':
            variables_in_order = variables
        elif order_variables == 'c':
            variables_in_order = np.asarray(variables_in_order, dtype=int)
            # check requirement 2
            if len(variables_in_order) != n_variables:
                raise ValueError('The following condition must be met:\nlen(variables_in_order) == len(conditions).')
            # check requirement 3
            for v in range(n_variables):
                if v not in variables_in_order:
                    raise ValueError('variables_in_order must contain all integers from 0 to (len(conditions)-1).\n'
                                     '{} is not in variables_in_order.'.format(v))
            # check requirement 4
            for v in variables_in_order:
                if np.sum(v == variables_in_order) > 1:
                    raise ValueError('variables_in_order cannot contain repeated integers.\n'
                                     '{} is a repeated integer.'.format(v))
        else:
            # check requirement 1
            raise ValueError(
                'order_variables can either be \'rl\' for right-to-left, '
                '\'lr\' for left-to-right or \'c\' for custom.\n'
                'Now, order_variables = {}'.format(order_variables))

        i_variable = variables_in_order[0]
        indexes_combinations = np.empty(2, dtype=object)
        indexes_combinations[axis_variables] = i_variable
        for i_condition in range(n_conditions[i_variable]):
            indexes_combinations[axis_combinations] = slice(i_condition, n_combinations, n_conditions[i_variable])
            combinations[tuple(indexes_combinations)] = conditions[i_variable][i_condition]

        indexes_combinations[axis_combinations] = slice(None)

        # combinations[tuple(indexes_combinations)] = np.expand_dims(
        #     cc_array.pad_array_from_n_samples_target(conditions[-1], n_samples_target=n_combinations),
        #     axis=axis_variables)
        cumulative_n_combinations = n_conditions[i_variable]
        for i_variable in variables_in_order[1:]:
            cumulative_combinations = np.empty(
                cumulative_n_combinations * n_conditions[i_variable], combinations.dtype)
            for i_condition in range(n_conditions[i_variable]):
                cumulative_combinations[
                    slice(i_condition * cumulative_n_combinations,
                          (i_condition + 1) * cumulative_n_combinations)] = conditions[i_variable][i_condition]

            indexes_combinations[axis_variables] = i_variable
            if cumulative_combinations.size < n_combinations:
                combinations[tuple(indexes_combinations)] = cc_array.pad_array_from_n_samples_target(
                    cumulative_combinations, n_samples_target=n_combinations)
            elif cumulative_combinations.size == n_combinations:
                combinations[tuple(indexes_combinations)] = cumulative_combinations
            else:
                raise Exception('something wrong')
            cumulative_n_combinations *= n_conditions[i_variable]

        if change_dtype:
            combinations = combinations.astype(dtype_end)
    else:
        combinations = []
    return combinations


def make_combinations_of_conditions_as_distributions(conditions_as_distributions, n_repetitions_combinations=1):

    n_variables = len(conditions_as_distributions)
    n_conditions = np.empty(n_variables, dtype=object)

    for i_variable in range(n_variables):

        n_conditions[i_variable] = len(conditions_as_distributions[i_variable])

    combinations_of_conditions = n_conditions_to_combinations(n_conditions)

    n_combinations = len(combinations_of_conditions)

    combinations_of_conditions_as_distributions = np.empty([n_combinations, n_variables], dtype=int)

    for i_variable in range(n_variables):

        for i_condition in range(n_conditions[i_variable]):

            indexes_of_i_condition = np.argwhere(combinations_of_conditions[:, i_variable] == i_condition)

            n_i_condition = len(indexes_of_i_condition)

            combinations_of_conditions_as_distributions[indexes_of_i_condition, i_variable] = \
                np.random.choice(conditions_as_distributions[i_variable][i_condition],
                                 n_i_condition)[:, None]

    return combinations_of_conditions_as_distributions


def trials_to_combinations(
        trials, axis_combinations=0, variables=None,  n_repetitions_combinations=1, dtype=None):

    conditions = trials_to_conditions(trials, axis_combinations=axis_combinations, variables=variables)
    combinations = conditions_to_combinations(
        conditions,
        axis_combinations=axis_combinations,
        n_repetitions_combinations=n_repetitions_combinations,
        dtype=dtype)
    return combinations


def trials_to_conditions(trials, axis_combinations=0, variables=None):

    axis_variables = int(not(bool(axis_combinations)))
    if variables is None:
        n_variables = trials.shape[axis_variables]
        variables = np.arange(n_variables)
    else:
        try:
            n_variables = len(variables)
        except TypeError:
            variables = np.asarray([variables], dtype=int)
            n_variables = len(variables)

    conditions = np.empty(n_variables, dtype=object)

    indexes_trials = np.empty(2, dtype=object)
    indexes_trials[axis_combinations] = slice(0, trials.shape[axis_combinations], 1)

    for i_variable in range(n_variables):

        indexes_trials[axis_variables] = variables[i_variable]
        trials_variables_i = trials[tuple(indexes_trials)]

        conditions[i_variable] = np.unique(trials_variables_i)

    return conditions


def trials_to_n_conditions(trials, axis_combinations=0, variables=None):
    conditions = trials_to_conditions(trials, axis_combinations=axis_combinations, variables=variables)
    n_conditions = conditions_to_n_conditions(conditions)
    # n_variables = len(conditions)
    # n_conditions = np.empty(n_variables, dtype=int)
    # for i_variable in range(n_variables):
    #     n_conditions[i_variable] = len(conditions[i_variable])
    return n_conditions


def n_conditions_to_conditions(n_conditions):
    n_variables = len(n_conditions)
    conditions = np.empty(n_variables, dtype=object)
    for i_variable in range(n_variables):
        conditions[i_variable] = np.arange(n_conditions[i_variable])
    return conditions


def conditions_to_n_conditions(conditions):
    n_variables = len(conditions)
    n_conditions = np.empty(n_variables, dtype=int)
    for i_variable in range(n_variables):
        n_conditions[i_variable] = len(conditions[i_variable])
    return n_conditions


def conditions_to_phases_of_blocks_of_shuffled_combinations(
        conditions, n_repetitions_combinations=1, n_repetitions_variables_repeated_per_block=1,
        variables_split=None, n_repetitions_variables_repeated_per_shuffle=None):

    n_phases = len(conditions)

    dict_arguments = dict(
        n_repetitions_combinations=n_repetitions_combinations,
        n_repetitions_variables_repeated_per_block=n_repetitions_variables_repeated_per_block,
        variables_split=variables_split,
        n_repetitions_variables_repeated_per_shuffle=n_repetitions_variables_repeated_per_shuffle)

    dict_arguments = cc_format.format_shape_arguments(dict_arguments, n_phases)
    n_repetitions_combinations = dict_arguments['n_repetitions_combinations']
    n_repetitions_variables_repeated_per_block = dict_arguments['n_repetitions_variables_repeated_per_block']
    variables_split = dict_arguments['variables_split']
    n_repetitions_variables_repeated_per_shuffle = dict_arguments['n_repetitions_variables_repeated_per_shuffle']

    # dict_arguments = cc_format.format_shape_arguments(dict_arguments, n_phases)
    # locals().update(dict_arguments)

    phases_of_blocks_of_shuffled_combinations = np.empty(n_phases, dtype=object)

    for i_phase in range(n_phases):

        phases_of_blocks_of_shuffled_combinations[i_phase] = conditions_to_blocks_of_shuffled_combinations(
            conditions[i_phase],
            n_repetitions_combinations=n_repetitions_combinations[i_phase],
            n_repetitions_variables_repeated_per_block=n_repetitions_variables_repeated_per_block[i_phase],
            n_repetitions_variables_repeated_per_shuffle=n_repetitions_variables_repeated_per_shuffle[i_phase],
            variables_split=variables_split[i_phase])

    return phases_of_blocks_of_shuffled_combinations


def n_conditions_to_phases_of_blocks_of_shuffled_combinations(
        n_conditions, n_repetitions_combinations=1, n_repetitions_variables_repeated_per_block=1,
        variables_split=None, n_repetitions_variables_repeated_per_shuffle=None):

    n_phases = len(n_conditions)

    dict_arguments = dict(
        n_repetitions_combinations=n_repetitions_combinations,
        n_repetitions_variables_repeated_per_block=n_repetitions_variables_repeated_per_block,
        variables_split=variables_split,
        n_repetitions_variables_repeated_per_shuffle=n_repetitions_variables_repeated_per_shuffle)

    dict_arguments = cc_format.format_shape_arguments(dict_arguments, n_phases)

    n_repetitions_combinations = dict_arguments['n_repetitions_combinations']
    n_repetitions_variables_repeated_per_block = dict_arguments['n_repetitions_variables_repeated_per_block']
    variables_split = dict_arguments['variables_split']
    n_repetitions_variables_repeated_per_shuffle = dict_arguments['n_repetitions_variables_repeated_per_shuffle']

    phases_of_blocks_of_shuffled_combinations = np.empty(n_phases, dtype=object)

    for i_phase in range(n_phases):
        phases_of_blocks_of_shuffled_combinations[i_phase] = n_conditions_to_blocks_of_shuffled_combinations(
            n_conditions[i_phase],
            n_repetitions_combinations=n_repetitions_combinations[i_phase],
            n_repetitions_variables_repeated_per_block=n_repetitions_variables_repeated_per_block[i_phase],
            n_repetitions_variables_repeated_per_shuffle=n_repetitions_variables_repeated_per_shuffle[i_phase],
            variables_split=variables_split[i_phase])

    return phases_of_blocks_of_shuffled_combinations


def n_conditions_to_blocks_of_shuffled_combinations(
        n_conditions, n_repetitions_combinations=1,
        n_repetitions_variables_repeated_per_block=1,
        n_repetitions_variables_repeated_per_shuffle=None,
        variables_split=None):

    conditions = n_conditions_to_conditions(n_conditions)

    trials_blocks = conditions_to_blocks_of_shuffled_combinations(
        conditions, n_repetitions_combinations=n_repetitions_combinations,
        n_repetitions_variables_repeated_per_block=n_repetitions_variables_repeated_per_block,
        n_repetitions_variables_repeated_per_shuffle=n_repetitions_variables_repeated_per_shuffle,
        variables_split=variables_split)

    return trials_blocks


def conditions_to_blocks_of_shuffled_combinations(
        conditions, n_repetitions_combinations=1,
        n_repetitions_variables_repeated_per_block=1,
        n_repetitions_variables_repeated_per_shuffle=None,
        variables_split=None):

    if type(conditions) != np.ndarray:
        conditions = np.array(conditions, dtype=object)

    n_conditions = conditions_to_n_conditions(conditions)

    axis_combinations = 0

    trials = conditions_to_combinations(
        conditions, axis_combinations=axis_combinations,
        n_repetitions_combinations=n_repetitions_combinations, order_variables='rl',
        variables_in_order=None, dtype=None)

    n_trials = trials.shape[axis_combinations]

    n_variables = conditions.shape[0]
    variables = np.arange(n_variables)

    try:
        n_variables_split = len(variables_split)
        if not isinstance(variables_split, np.ndarray):
            variables_split = np.asarray(variables_split, dtype=int)
    except TypeError:
        if variables_split is None:
            n_variables_split = 0
        else:
            variables_split = np.asarray([variables_split], dtype=int)
            n_variables_split = 1

    if n_variables_split == 0:

        conditions_variables_repeated = conditions
        n_conditions_variables_repeated = n_conditions
        if n_repetitions_variables_repeated_per_shuffle is None:
            n_repetitions_variables_repeated_per_shuffle = n_repetitions_variables_repeated_per_block
    else:
        variables_split[variables_split < 0] += n_variables
        n_repetitions_variables_repeated_per_shuffle = n_repetitions_variables_repeated_per_block

        if np.any(np.logical_or(variables_split < 0, variables_split >= n_variables)):
            raise Exception('There is at least 1 variables_split that is not in the variables.\n'
                            'each sample in the variables_split has to be equal to either \n'
                            'one or more samples of variables or None.\n'
                            'Now, variables_split = {} and variables = {}.\n'
                            'By default, variables_split=None.'.format(
                                variables_split, variables))

        variables_repeated = np.where([i not in variables_split for i in variables])[0]
        conditions_variables_repeated = conditions[variables_repeated]
        n_conditions_variables_repeated = n_conditions[variables_repeated]

        conditions_variables_split = conditions[variables_split]
        n_conditions_variables_split = n_conditions[variables_split]
        n_combinations_variables_split = cc_maths.prod(n_conditions_variables_split)

    n_combinations_variables_repeated = cc_maths.prod(n_conditions_variables_repeated)

    n_trials_variables_repeated = n_combinations_variables_repeated * n_repetitions_combinations

    n_trials_per_block = n_combinations_variables_repeated * n_repetitions_variables_repeated_per_block

    if n_trials % n_trials_per_block != 0:
        possibles_n_repetitions_variables_repeated_per_block = cc_maths.factors_of_x(
            n_trials // n_combinations_variables_repeated)

        raise Exception(
            'n_trials_per_blocks are not equal. Change the n_repetitions_variables_repeated_per_block\n'
            'or the n_repetitions_combinations to make the n_trials_per_blocks equal.\n'
            'Given n_repetitions_combinations = {} and n_combinations_variables_repeated = {},\n'
            'n_trials = {} and n_repetitions_variables_repeated_per_block can either\n'
            'be one of the following numbers {}. Now, n_repetitions_variables_repeated_per_block = {}.'.format(
                n_repetitions_combinations, n_combinations_variables_repeated, n_trials,
                possibles_n_repetitions_variables_repeated_per_block,
                n_repetitions_variables_repeated_per_block))

    n_blocks = n_trials // n_trials_per_block

    n_trials_per_shuffle = n_combinations_variables_repeated * n_repetitions_variables_repeated_per_shuffle

    if n_trials % n_trials_per_shuffle != 0:

        possibles_n_repetitions_variables_repeated_per_shuffle = cc_maths.factors_of_x(
            n_trials // n_combinations_variables_repeated)

        raise Exception(
            'n_trials_per_shuffles are not equal. Change the n_repetitions_variables_repeated_per_shuffle\n'
            'or the n_repetitions_combinations to make the n_trials_per_shuffles equal.\n'
            'Given n_repetitions_combinations = {} and n_combinations_variables_repeated = {},\n'
            'n_trials = {} and n_repetitions_variables_repeated_per_shuffle can either\n'
            'be one of the following numbers {}. Now, n_repetitions_variables_repeated_per_shuffle = {}.'.format(
                n_repetitions_combinations, n_combinations_variables_repeated, n_trials,
                possibles_n_repetitions_variables_repeated_per_shuffle,
                n_repetitions_variables_repeated_per_shuffle))

    n_shuffles = int(n_trials / n_trials_per_shuffle)

    if n_variables_split == 0:

        trials_blocks = np.split(trials, n_blocks, axis=axis_combinations)

    else:

        trials_1_repetition_variables_repeated = conditions_to_combinations(
            conditions_variables_repeated, axis_combinations=0, n_repetitions_combinations=1,
            order_variables='rl', variables_in_order=None, dtype=None)

        trials_1_repetition_variables_split = conditions_to_combinations(
            conditions_variables_split, axis_combinations=0, n_repetitions_combinations=1,
            order_variables='rl', variables_in_order=None, dtype=None)

        range_combinations_variables_repeated = np.arange(n_combinations_variables_repeated)

        min_trials_of_1_single_variable_split_per_block = np.empty(n_variables_split, dtype=int)
        available_conditions_variables_split_per_block = np.empty(n_variables_split, dtype=object)
        extra_available_conditions_variables_split = np.empty(n_variables_split, dtype=object)
        n_available_conditions_variables_split_per_block = np.empty(n_variables_split, dtype=int)

        for i_variable_split in range(n_variables_split):

            n_conditions_variables_split_not_in_variables_split = n_conditions_variables_split[
                np.arange(n_variables_split) != i_variable_split]

            min_trials_of_1_single_variable_split_per_block[i_variable_split] = \
                (n_combinations_variables_repeated * cc_maths.prod(
                    n_conditions_variables_split_not_in_variables_split) *
                 n_repetitions_combinations) // n_blocks

            available_conditions_variables_split_per_block[i_variable_split] = np.arange(
                n_conditions_variables_split[i_variable_split] *
                min_trials_of_1_single_variable_split_per_block[i_variable_split]) // \
                min_trials_of_1_single_variable_split_per_block[i_variable_split]

            extra_available_conditions_variables_split[i_variable_split] = np.arange(
                n_conditions_variables_split[i_variable_split])

            n_available_conditions_variables_split_per_block[i_variable_split] = len(
                available_conditions_variables_split_per_block[i_variable_split])

        n_extra_conditions_variables_split = n_trials_per_block % n_conditions_variables_split

        i_trial_block = np.empty(shape=[1, n_variables], dtype=int)  # TO MOVE BEFORE THE FOR LOOPS
        # k = 0
        u = 0
        block = True
        while block:
            u += 1
            print('Block {}'.format(u))

            block = False

            tmp_trials = trials
            shape_frequencies = n_conditions_variables_split
            frequencies_combinations_repeated_and_split = np.zeros(
                (n_combinations_variables_repeated, *n_conditions_variables_split), dtype=int)

            extra_available_conditions_variables_split_per_i_block = np.copy(extra_available_conditions_variables_split)

            available_conditions_variables_split_per_i_block = np.empty(n_blocks, object)
            trials_blocks = np.empty(n_blocks, dtype=object)

            for i_block in range(n_blocks):

                available_conditions_variables_split_per_i_block[i_block] = np.copy(
                    available_conditions_variables_split_per_block)

                n_available_conditions_variables_split_per_i_block = np.empty(n_variables_split, dtype=int)

                for i_variable_split in range(n_variables_split):

                    if n_extra_conditions_variables_split[i_variable_split] > 0:

                        for i_extra_condition_variables_split in range(
                                n_extra_conditions_variables_split[i_variable_split]):

                            n_extra_available_conditions_variables_split_per_i_block = len(
                                extra_available_conditions_variables_split_per_i_block[i_variable_split])

                            if n_extra_available_conditions_variables_split_per_i_block < 1:
                                extra_available_conditions_variables_split_per_i_block[i_variable_split] = \
                                    np.copy(extra_available_conditions_variables_split[i_variable_split])

                            extra_available_conditions_variables_split_per_i_block[i_variable_split], \
                                available_conditions_variables_split_per_i_block[i_block][i_variable_split] = \
                                cc_array.transfer_n_random_samples_from_arr1_to_arr2(
                                    extra_available_conditions_variables_split_per_i_block[i_variable_split],
                                    available_conditions_variables_split_per_i_block[i_block][i_variable_split],
                                    n_samples=1, axis=0, replace=False)

                    n_available_conditions_variables_split_per_i_block[i_variable_split] = len(
                        available_conditions_variables_split_per_i_block[i_block][i_variable_split])

                if np.any(n_available_conditions_variables_split_per_i_block[0] !=
                          n_available_conditions_variables_split_per_i_block):
                    raise Warning(
                        'lengths of n_available_conditions_variables_split_per_i_block are not all equal.\n'
                        'n_available_conditions_variables_split_per_i_block = {}'.format(
                            n_available_conditions_variables_split_per_block))

                trials_blocks[i_block] = np.empty((0, n_variables), int)

                for i_combinations_variables_repeated in range(n_combinations_variables_repeated):

                    i_repetition_variables_repeated_per_block = 0
                    while i_repetition_variables_repeated_per_block < n_repetitions_variables_repeated_per_block:

                        available_combinations_variables_split_per_i_trial = np.argwhere(
                            frequencies_combinations_repeated_and_split[
                                i_combinations_variables_repeated] <
                            n_repetitions_combinations)

                        tmp_logical_index_available_combinations_variables_split_per_i_block_per_i_trial = np.empty(
                            (len(available_combinations_variables_split_per_i_trial), n_variables_split), dtype=bool)

                        for i_variable_split in range(n_variables_split):
                            tmp_logical_index_available_combinations_variables_split_per_i_block_per_i_trial[
                                slice(None), i_variable_split] = cc_array.samples_in_arr1_are_in_arr2(
                                    available_combinations_variables_split_per_i_trial[:, i_variable_split],
                                    available_conditions_variables_split_per_i_block[i_block][i_variable_split])

                        logical_indexes_available_combinations_variables_split_per_i_block_per_i_trial = np.all(
                            tmp_logical_index_available_combinations_variables_split_per_i_block_per_i_trial, axis=1)

                        available_combinations_variables_split_per_i_block_per_i_trial = \
                            available_combinations_variables_split_per_i_trial[
                                logical_indexes_available_combinations_variables_split_per_i_block_per_i_trial]

                        n_available_combinations_variables_split_per_i_block_per_i_trial = len(
                            available_combinations_variables_split_per_i_block_per_i_trial)

                        if n_available_combinations_variables_split_per_i_block_per_i_trial > 0:

                            index_of_chosen_combination_variables_split = np.random.randint(
                                n_available_combinations_variables_split_per_i_block_per_i_trial)

                            chosen_combination_variables_split = np.expand_dims(
                                available_combinations_variables_split_per_i_block_per_i_trial[
                                    index_of_chosen_combination_variables_split], axis=0)

                            i_trial_block[0, variables_repeated] = trials_1_repetition_variables_repeated[
                                i_combinations_variables_repeated]

                            for i_variable_split in range(n_variables_split):
                                i_trial_block[0, variables_split[i_variable_split]] = conditions_variables_split[i_variable_split][
                                    chosen_combination_variables_split[0, i_variable_split]]

                            trials_blocks[i_block] = np.append(trials_blocks[i_block], i_trial_block, axis=0)

                            indexes_frequencies_combinations = np.empty(n_variables_split + 1, object)
                            for i_variable_split in range(n_variables_split + 1):
                                if i_variable_split == 0:
                                    indexes_frequencies_combinations[i_variable_split] = np.asarray(
                                        [i_combinations_variables_repeated])
                                else:
                                    indexes_frequencies_combinations[i_variable_split] = np.asarray(
                                        [chosen_combination_variables_split[0, i_variable_split - 1]])

                            frequencies_combinations_repeated_and_split[np.ix_(*indexes_frequencies_combinations)] += 1

                            for i_variable_split in range(n_variables_split):

                                first_index_chosen_conditions_variables_split = np.argwhere(
                                    available_conditions_variables_split_per_i_block[i_block][i_variable_split] == \
                                    chosen_combination_variables_split[0, i_variable_split])[0]

                                available_conditions_variables_split_per_i_block[i_block][i_variable_split] = np.delete(
                                    available_conditions_variables_split_per_i_block[i_block][i_variable_split],
                                    first_index_chosen_conditions_variables_split,  axis=0)

                            logical_index_i_trial = np.all(tmp_trials == i_trial_block, axis=1)
                            index_i_trial = np.argwhere(logical_index_i_trial)[0]

                            tmp_trials = np.delete(tmp_trials, index_i_trial, 0)
                            i_repetition_variables_repeated_per_block += 1

                        else:
                            # which trial in this block I can put the last actors available for this block that are not
                            # available for this trial

                            tmp_conditions_variables_split = np.empty(n_variables_split, dtype=object)
                            tmp_extra_conditions_variables_split = np.empty(n_variables_split, dtype=object)
                            j = 0
                            i = 0
                            for i_variable_split in range(n_variables_split):

                                if len(available_conditions_variables_split_per_i_block[
                                        i_block][i_variable_split]) > 0:

                                    tmp_conditions_variables_split[j] = np.unique(
                                        available_conditions_variables_split_per_i_block[i_block][i_variable_split])
                                    j += 1
                                else:
                                    tmp_conditions_variables_split = np.delete(
                                        tmp_conditions_variables_split, i_variable_split, axis=0)

                                if len(extra_available_conditions_variables_split_per_i_block[i_variable_split]) > 0:

                                    tmp_extra_conditions_variables_split[i] = \
                                        extra_available_conditions_variables_split_per_i_block[i_variable_split]
                                    i += 1
                                else:
                                    tmp_extra_conditions_variables_split = np.delete(
                                        tmp_extra_conditions_variables_split, i, axis=0)

                            if len(tmp_conditions_variables_split) > 1:
                                tmp_combinations_variables_split = conditions_to_combinations(
                                    tmp_conditions_variables_split, axis_combinations=0,
                                    n_repetitions_combinations=1, order_variables='rl',
                                    variables_in_order=None, dtype=None)
                            else:
                                tmp_combinations_variables_split = np.array(list(
                                    tmp_conditions_variables_split), dtype=int)

                            if len(tmp_extra_conditions_variables_split) > 1:
                                tmp_combinations_extra_conditions_variables_split = (
                                    conditions_to_combinations(
                                        tmp_extra_conditions_variables_split, axis_combinations=0,
                                        n_repetitions_combinations=1, order_variables='rl',
                                        variables_in_order=None, dtype=None))
                            else:
                                tmp_combinations_extra_conditions_variables_split = np.array(
                                    list(tmp_extra_conditions_variables_split), dtype=int)

                            n_tries = 2
                            try_available_combinations_variables_split_for_i_block = np.empty(
                                n_tries, dtype=object)

                            try_available_combinations_variables_split_for_i_block[0] = \
                                np.random.permutation(tmp_combinations_variables_split)

                            try_available_combinations_variables_split_for_i_block[1] = \
                                np.random.permutation(tmp_combinations_extra_conditions_variables_split)

                            try_available_combinations_variables_split_for_i_trial = np.random.permutation(
                                available_combinations_variables_split_per_i_trial)

                            n_available_combinations_variables_split_for_i_trial = len(
                                available_combinations_variables_split_per_i_trial)

                            search_break = False

                            for i_try in range(n_tries):

                                n_try_available_combinations_variables_split_for_i_block = \
                                    try_available_combinations_variables_split_for_i_block[i_try].shape[0]

                                for i_combinations_variables_split_for_i_block in range(
                                        n_try_available_combinations_variables_split_for_i_block):

                                    for i_combinations_variables_split_for_i_trial in range(
                                            n_available_combinations_variables_split_for_i_trial):

                                        indexes_frequencies_combinations = np.empty(n_variables_split + 1, object)
                                        for i_variable_split in range(n_variables_split + 1):
                                            if i_variable_split == 0:
                                                indexes_frequencies_combinations[i_variable_split] = \
                                                    range_combinations_variables_repeated
                                            else:
                                                indexes_frequencies_combinations[i_variable_split] = np.asarray(
                                                    [try_available_combinations_variables_split_for_i_block[
                                                    i_try][i_combinations_variables_split_for_i_block,
                                                           i_variable_split - 1]])

                                        logical_indexes_available_combinations_variables_repeated_for_i_try_compination_split_from_i_shaffle = np.squeeze(
                                            frequencies_combinations_repeated_and_split[np.ix_(*indexes_frequencies_combinations)] <
                                            n_repetitions_combinations)

                                        available_combinations_variables_repeated_for_i_try_compination_split_from_i_shaffle = trials_1_repetition_variables_repeated[
                                            logical_indexes_available_combinations_variables_repeated_for_i_try_compination_split_from_i_shaffle]

                                        n_available_combinations_variables_repeated_for_i_try_compination_split_from_i_shaffle = available_combinations_variables_repeated_for_i_try_compination_split_from_i_shaffle.shape[0]

                                        n_2 = trials_blocks[i_block].shape[0]
                                        log2 = np.empty(shape=(n_2, n_available_combinations_variables_repeated_for_i_try_compination_split_from_i_shaffle), dtype=bool)
                                        for l in range(n_available_combinations_variables_repeated_for_i_try_compination_split_from_i_shaffle):
                                            log2[:, l] = np.all(
                                                trials_blocks[i_block][:, variables_repeated] ==
                                                available_combinations_variables_repeated_for_i_try_compination_split_from_i_shaffle[l, :], axis=1)

                                        log3 = np.any(log2, axis=1)

                                        inx = np.empty(n_variables_split, dtype=int)
                                        for i_variable_split in range(n_variables_split):
                                            inx[i_variable_split] = (
                                                conditions_variables_split[i_variable_split][
                                                    try_available_combinations_variables_split_for_i_trial[
                                                        i_combinations_variables_split_for_i_trial, i_variable_split]])

                                        log4 = np.all(trials_blocks[i_block][:, variables_split] == inx, axis=1)

                                        a = try_available_combinations_variables_split_for_i_trial[
                                            i_combinations_variables_split_for_i_trial]

                                        log5 = np.all(np.array([log3, log4]), axis=0)

                                        if np.sum(log5) > 0:

                                            inx2 = np.empty(n_variables_split, dtype=int)
                                            for i_variable_split in range(n_variables_split):
                                                inx2[i_variable_split] = conditions_variables_split[
                                                    i_variable_split][
                                                    try_available_combinations_variables_split_for_i_block[
                                                        i_try][i_combinations_variables_split_for_i_block, i_variable_split]]

                                            chosen_combination_variables_split_from_i_block = inx2
                                            indexes_chosen_combination_variables_split_from_i_block = \
                                                try_available_combinations_variables_split_for_i_block[
                                                    i_try][i_combinations_variables_split_for_i_block]

                                            chosen_combination_variables_split_for_i_trial = inx
                                            indexes_chosen_combination_variables_split_for_i_trial = \
                                                try_available_combinations_variables_split_for_i_trial[
                                                    i_combinations_variables_split_for_i_trial]

                                            if i_try == 0:
                                                search_break = True
                                                break

                                            elif i_try == 1:

                                                try:

                                                    conditions_variables_split_leaving_i_block = np.empty(n_variables_split, dtype=int)

                                                    for i_variable_split in range(n_variables_split):

                                                        conditions_variables_split_leaving_i_block[i_variable_split] = \
                                                            np.random.choice(np.setdiff1d(
                                                                available_conditions_variables_split_per_i_block[
                                                                    i_block][i_variable_split],
                                                                extra_available_conditions_variables_split_per_i_block[
                                                                    i_variable_split]), 1)

                                                        first_index_leaving = np.argwhere(
                                                            available_conditions_variables_split_per_i_block[
                                                                i_block][i_variable_split]
                                                            == conditions_variables_split_leaving_i_block[i_variable_split])[0, 0]

                                                        available_conditions_variables_split_per_i_block[
                                                            i_block][i_variable_split][first_index_leaving] = \
                                                            indexes_chosen_combination_variables_split_from_i_block[
                                                                i_variable_split]

                                                        extra_available_conditions_variables_split_per_i_block[
                                                            i_variable_split][
                                                            extra_available_conditions_variables_split_per_i_block[
                                                            i_variable_split] ==
                                                            indexes_chosen_combination_variables_split_from_i_block[
                                                                i_variable_split]] = conditions_variables_split_leaving_i_block

                                                except ValueError:
                                                    pass

                                        if i_try == n_tries - 1:
                                            if i_combinations_variables_split_for_i_block == \
                                                    n_try_available_combinations_variables_split_for_i_block - 1 \
                                                    and i_combinations_variables_split_for_i_trial == \
                                                    n_available_combinations_variables_split_for_i_trial - 1:
                                                block = True
                                                search_break = True
                                                break



                                    if search_break:
                                        break

                                if search_break:
                                    search_break = False
                                    break

                            if block:
                                break

                            c = np.reshape(np.argwhere(log5), (-1))


                            chosen_trial = np.random.choice(c, 1)

                            trials_blocks[i_block][
                                chosen_trial, variables_split] = chosen_combination_variables_split_from_i_block


                            ddds = np.argwhere(np.all(
                                tmp_trials == trials_blocks[i_block][chosen_trial, :],
                                axis=1))[0]
                            #

                            tmp_trials[ddds, variables_split] = chosen_combination_variables_split_for_i_trial

                            index_l = np.all(
                                trials_blocks[i_block][chosen_trial, variables_repeated] ==
                                trials_1_repetition_variables_repeated, axis=1)

                            if np.sum(index_l) != 1:
                                raise Exception(
                                    'sum of index_l should either be 1. Now it was {}.'.format(np.sum(index_l)))
                            index_l = np.argwhere(index_l)[0]

                            indexes_frequencies_combinations = np.empty(n_variables_split + 1, object)
                            for i_variable_split in range(n_variables_split + 1):
                                if i_variable_split == 0:
                                    indexes_frequencies_combinations[i_variable_split] = index_l
                                else:
                                    indexes_frequencies_combinations[i_variable_split] = np.asarray(
                                        [indexes_chosen_combination_variables_split_for_i_trial[
                                            i_variable_split - 1]])

                            frequencies_combinations_repeated_and_split[np.ix_(*indexes_frequencies_combinations)] -= 1

                            indexes_frequencies_combinations = np.empty(n_variables_split + 1, object)
                            for i_variable_split in range(n_variables_split + 1):
                                if i_variable_split == 0:
                                    indexes_frequencies_combinations[i_variable_split] = index_l
                                else:
                                    indexes_frequencies_combinations[i_variable_split] = np.asarray(
                                        [indexes_chosen_combination_variables_split_from_i_block[
                                            i_variable_split - 1]])

                            frequencies_combinations_repeated_and_split[np.ix_(*indexes_frequencies_combinations)] += 1

                            for i_variable_split in range(n_variables_split):

                                first_index_chosen_conditions_variables_split = np.argwhere(
                                    available_conditions_variables_split_per_i_block[i_block][i_variable_split] ==
                                    indexes_chosen_combination_variables_split_from_i_block[i_variable_split])[0]

                                available_conditions_variables_split_per_i_block[i_block][i_variable_split][
                                    first_index_chosen_conditions_variables_split] = \
                                    indexes_chosen_combination_variables_split_for_i_trial[i_variable_split]

                    if block:
                        break

                if block:
                    break

                if len(trials_blocks[i_block]) != n_trials_per_block:
                    raise Exception('the n trials in trials_blocks[i_block] should be equal to n_trials_per_block. '
                                    'n trials in trials_blocks[i_block] was {} whereas n_trials_per_block was {}'
                                    '.'.format(len(trials_blocks[i_block]), n_trials_per_block))
                else:
                    pass

        if np.sum(frequencies_combinations_repeated_and_split) != n_trials:

            raise Exception('At the end, np.sum(frequencies_combinations_repeated_and_split should be n_trials.\n'
                            'Now, np.sum(frequencies_combinations_repeated_and_split) = {}\n'
                            'and n_trials = {}.'.format(np.sum(frequencies_combinations_repeated_and_split),
                                                        n_trials))

        n_unused_trials = tmp_trials.shape[0]
        if n_unused_trials != 0:
            raise Exception(
                'n_unused_trials should be 0. From tmp_trials, n_unused_trials was {}.'.format(n_unused_trials))

    for i_block in range(n_blocks):
        trials_blocks[i_block] = cc_array.shuffle_in_windows(
            trials_blocks[i_block], n_samples_window=n_trials_per_shuffle,
            n_windows=None, axis=axis_combinations, dtype=None)

    return trials_blocks


def conditions_to_permutetions(conditions, R, dtype=None):
    permutations = np.asarray(list(itertools.permutations(conditions, R)), dtype=dtype)
    return permutations


def conditions_to_P_random_permutetions(conditions, R=None, P=None, dtype=None):

    C = len(conditions)
    if R == C or R is None:
        n = math.factorial(C)
    else:
        n = cc_maths.convert_to_int_or_float(math.factorial(C) / math.factorial(C - R))

    if P is None or P >= n:
        permutations = conditions_to_permutetions(conditions, R)
        permutations = cc_array.shuffle_in_any_dimension(permutations, 0)
        return permutations

    else:
        indexes_cond = np.arange(C)
        positions = np.arange(R)

        permutations_indexes = conditions_to_permutetions(indexes_cond, R, dtype='i')
        permutations_indexes = cc_array.shuffle_in_any_dimension(permutations_indexes, 0)

        max_per_pos = math.ceil(20 / C)
        frequencies = np.empty([C, R], dtype='i')
        frequencies[:] = 0

        indexes_perm = np.empty(P, dtype='i')
        p = 0
        for i in range(n):
            if np.all(frequencies[tuple([indexes_cond[permutations_indexes[i, :]], positions])] < max_per_pos):
                frequencies[tuple([indexes_cond[permutations_indexes[i, :]], positions])] += 1
                indexes_perm[p] = p
                p += 1
                if p == P:
                    break

        if p == P:
            permutations_indexes = permutations_indexes[indexes_perm, :]
        elif 0 < p < P:
            indexes_perm = indexes_perm[:p]
            permutations_indexes = permutations_indexes[indexes_perm, :]
        else:
            permutations = None
            return permutations

        if np.equal(indexes_cond, conditions).all():
            return permutations_indexes
        else:
            if dtype is None:
                if isinstance(conditions, np.ndarray):
                    dtype = conditions.dtype
                else:
                    dtype = np.asarray(conditions).dtype
            permutations = np.empty([P, R], dtype=dtype)
            for c in range(C):
                permutations[permutations_indexes == c] = conditions[c]

            return permutations





import numpy as np
from ..combinations import n_conditions_to_combinations, conditions_to_combinations
from ..array import samples_in_arr1_are_in_arr2
from ..format import seq_of_numeric_indexes_to_seq_of_slices


def merge_axes(
        array_input,
        axes_removing_input,
        axis_pooling_input,
        dtype=None):
    # Input requirements:
    # 1) axes_removing_input cannot contain same values;
    # 2) axes_removing_input[a] != axis_pooling_input

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # format axes_removing_input
    if isinstance(axes_removing_input, list) or isinstance(axes_removing_input, tuple):
        axes_removing_input = np.asarray(axes_removing_input, dtype=int)
        n_axes_removing_input = axes_removing_input.size
        if n_axes_removing_input == 1:
            axes_removing_input = axes_removing_input[0]
    elif isinstance(axes_removing_input, np.ndarray):
        if len(axes_removing_input.shape) == 0:
            n_axes_removing_input = 1
        else:
            n_axes_removing_input = axes_removing_input.size
            if n_axes_removing_input == 1:
                axes_removing_input = axes_removing_input[0]
    else:
        n_axes_removing_input = 1

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = shape_array_input.size
    if axis_pooling_input < 0:
        axis_pooling_input += n_axes_array_input
    axis_pooling_output = axis_pooling_input - np.sum(axes_removing_input < axis_pooling_input)

    # check point 1
    if n_axes_removing_input > 1:
        for a in axes_removing_input:
            if np.sum(a == axes_removing_input) > 1:
                raise ValueError('axes_removing_input cannot contain repeated values. {} is repeated.'.format(a))

    # check point 2
    if np.sum(axes_removing_input == axis_pooling_input) > 0:
        raise ValueError('The following condition is not met:\n'
                         '\taxes_removing_input[a] \u2260 axis_pooling_input')

    if n_axes_removing_input == 1:
        if axes_removing_input < 0:
            axes_removing_input += n_axes_array_input
        if axes_removing_input < 0 or axes_removing_input >= n_axes_array_input:
            raise np.AxisError(
                'axes_removing_input = {} is out of range for array_input of {} dimensions'.format(
                    axes_removing_input, n_axes_array_input))
        n_conditions_in_axes_removing_input = [shape_array_input[axes_removing_input]]
    elif n_axes_removing_input > 1:
        axes_removing_input[axes_removing_input < 0] += n_axes_array_input
        if np.any(axes_removing_input < 0) or np.any(axes_removing_input >= n_axes_array_input):
            raise np.AxisError(
                'axes_removing_input = {} is out of range for array_input of {} dimensions'.format(
                    axes_removing_input, n_axes_array_input))
        n_conditions_in_axes_removing_input = shape_array_input[axes_removing_input]
    
    axes_merging_input = np.append(axis_pooling_input, axes_removing_input)
    # n_axes_merging_input = axes_merging_input.size

    axes_array_input = np.arange(n_axes_array_input)
    axes_other_array_input = axes_array_input[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_array_input, axes_merging_input))]

    n_axes_array_output = n_axes_array_input - n_axes_removing_input
    axes_array_output = np.arange(n_axes_array_output)
    axes_other_output = axes_array_output[axes_array_output != axis_pooling_output]

    shape_array_output = np.empty(n_axes_array_output, dtype=int)
    n_conditions_in_axes_merging_input = shape_array_input[axes_merging_input]
    shape_array_output[axis_pooling_output] = np.prod(n_conditions_in_axes_merging_input)
    shape_array_output[axes_other_output] = shape_array_input[axes_other_array_input]
    if dtype is None:
        dtype = array_input.dtype
    array_output = np.empty(shape_array_output, dtype=dtype)

    axis_combinations = 0
    axis_variables = 1
    combinations_axes_removing_input = n_conditions_to_combinations(
        n_conditions_in_axes_removing_input, axis_combinations=axis_combinations, order_variables='lr')
    n_combinations_axes_removing_input = combinations_axes_removing_input.shape[axis_combinations]

    indexes_combinations = np.empty(2, dtype=object)
    if n_axes_removing_input == 1:
        indexes_combinations[axis_variables] = 0
    elif n_axes_removing_input > 1:
        indexes_combinations[axis_variables] = slice(0, n_axes_removing_input, 1)

    indexes_input = np.empty(n_axes_array_input, dtype=object)
    indexes_input[:] = slice(0, None, 1)

    indexes_output = np.empty(n_axes_array_output, dtype=object)
    indexes_output[:] = slice(0, None, 1)

    length_axis_pooling_input = shape_array_input[axis_pooling_input]
    start_index_axis_pooling_output = 0
    for c in range(n_combinations_axes_removing_input):
        stop_index_axis_pooling_output = (c + 1) * length_axis_pooling_input
        indexes_output[axis_pooling_output] = slice(
            start_index_axis_pooling_output, stop_index_axis_pooling_output)
        start_index_axis_pooling_output = stop_index_axis_pooling_output

        indexes_combinations[axis_combinations] = c
        indexes_input[axes_removing_input] = (
            combinations_axes_removing_input[tuple(indexes_combinations)])
        array_output[tuple(indexes_output)] = array_input[tuple(indexes_input)]

    return array_output


def merge_axes_and_add_axes_removing_info_to_axis_receiving(
        array_input,
        axes_removing_input,
        axis_pooling_input,
        axis_receiving_input,
        indexes_inserting_output=None,
        dtype=None):

    # Input requirements:
    # 1) axes_removing_input cannot contain same values;
    # 2) indexes_inserting_output cannot contain same values;
    # 3) if indexes_inserting_output is not None,
    #        shapes of axes_removing_input and indexes_inserting_output must be equal
    # 4) axes_removing_input[a] != axis_pooling_input != axis_receiving_input

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # format axes_removing_input
    if isinstance(axes_removing_input, list) or isinstance(axes_removing_input, tuple):
        axes_removing_input = np.asarray(axes_removing_input, dtype=int)
        n_axes_removing_input = axes_removing_input.size
        if n_axes_removing_input == 1:
            axes_removing_input = axes_removing_input[0]
    elif isinstance(axes_removing_input, np.ndarray):
        if len(axes_removing_input.shape) == 0:
            n_axes_removing_input = 1
        else:
            n_axes_removing_input = axes_removing_input.size
            if n_axes_removing_input == 1:
                axes_removing_input = axes_removing_input[0]
    else:
        n_axes_removing_input = 1

    # format indexes_inserting_output
    if indexes_inserting_output is None:
        n_indexes_inserting_output = n_axes_removing_input
        if n_indexes_inserting_output == 1:
            indexes_inserting_output = 0
        else:
            indexes_inserting_output = np.arange(0, n_indexes_inserting_output, 1, dtype=int)
    elif isinstance(indexes_inserting_output, list) or isinstance(indexes_inserting_output, tuple):
        indexes_inserting_output = np.asarray(indexes_inserting_output, dtype=int)
        n_indexes_inserting_output = indexes_inserting_output.size
        if n_indexes_inserting_output == 1:
            indexes_inserting_output = indexes_inserting_output[0]
    elif isinstance(indexes_inserting_output, np.ndarray):
        if len(indexes_inserting_output.shape) == 0:
            n_indexes_inserting_output = 1
        else:
            n_indexes_inserting_output = indexes_inserting_output.size
            if n_indexes_inserting_output == 1:
                indexes_inserting_output = indexes_inserting_output[0]
    else:
        n_indexes_inserting_output = 1

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = shape_array_input.size
    if axis_pooling_input < 0:
        axis_pooling_input += n_axes_array_input
    if axis_receiving_input < 0:
        axis_receiving_input += n_axes_array_input
    length_axis_receiving_input = shape_array_input[axis_receiving_input]
    # indexes_in_axis_receiving_input = np.arange(length_axis_receiving_input)
    length_axis_receiving_output = length_axis_receiving_input + n_indexes_inserting_output
    axis_receiving_output = axis_receiving_input - np.sum(axes_removing_input < axis_receiving_input)
    axis_pooling_output = axis_pooling_input - np.sum(axes_removing_input < axis_pooling_input)

    # check point 1
    if n_axes_removing_input > 1:
        for a in axes_removing_input:
            if np.sum(a == axes_removing_input) > 1:
                raise ValueError('axes_removing_input cannot contain repeated values. {} is repeated.'.format(a))

    # check point 2
    if n_indexes_inserting_output > 1:
        for i in indexes_inserting_output:
            if np.sum(i == indexes_inserting_output) > 1:
                raise ValueError('indexes_inserting_output cannot contain repeated values. {} is repeated.'.format(i))

    # check point 3
    if n_indexes_inserting_output != n_axes_removing_input:
        raise ValueError(
            'Shapes of axes_removing_input and indexes_inserting_output must be equal')

    # check point 4
    if np.sum(axes_removing_input == axis_pooling_input) > 0:
        raise ValueError('The following condition is not met:\n'
                         '\taxes_removing_input[a] \u2260 axis_pooling_input')
    if np.sum(axes_removing_input == axis_receiving_input) > 0:
        raise ValueError('The following condition is not met:\n'
                         '\taxes_removing_input[a] \u2260 axis_receiving_input')
    if axis_pooling_input == axis_receiving_input:
        raise ValueError('The following condition is not met:\n'
                         '\taxis_pooling_input \u2260 axis_receiving_input')

    if n_axes_removing_input == 1:
        if axes_removing_input < 0:
            axes_removing_input += n_axes_array_input
        if axes_removing_input < 0 or axes_removing_input >= n_axes_array_input:
            raise np.AxisError(
                'axes_removing_input = {} is out of range for array_input of {} dimensions'.format(
                    axes_removing_input, n_axes_array_input))
        n_conditions_in_axes_removing_input = [shape_array_input[axes_removing_input]]

        if indexes_inserting_output < 0:
            indexes_inserting_output += length_axis_receiving_output
        if indexes_inserting_output < 0 or indexes_inserting_output >= length_axis_receiving_output:
            raise IndexError(
                'indexes_inserting_output = {} is out of range for axis_receiving_output = {} with size = {}'.format(
                    indexes_inserting_output, axis_receiving_output, length_axis_receiving_output))
    elif n_axes_removing_input > 1:
        axes_removing_input[axes_removing_input < 0] += n_axes_array_input
        if np.any(axes_removing_input < 0) or np.any(axes_removing_input >= n_axes_array_input):
            raise np.AxisError(
                'axes_removing_input = {} is out of range for array_input of {} dimensions'.format(
                    axes_removing_input, n_axes_array_input))
        n_conditions_in_axes_removing_input = shape_array_input[axes_removing_input]

        indexes_inserting_output[indexes_inserting_output < 0] += length_axis_receiving_output
        if np.any(indexes_inserting_output < 0) or np.any(indexes_inserting_output >= length_axis_receiving_output):
            raise IndexError(
                'indexes_inserting_output = {} is out of range for axis_receiving_output = {} with size = {}'.format(
                    indexes_inserting_output, axis_receiving_output, length_axis_receiving_output))

    axes_merging_input = np.append(axis_pooling_input, axes_removing_input)
    # n_axes_merging_input = axes_merging_input.size

    axes_array_input = np.arange(n_axes_array_input)
    axes_other_array_input = axes_array_input[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_array_input, np.append(axes_merging_input, axis_receiving_input)))]

    n_axes_array_output = n_axes_array_input - n_axes_removing_input
    axes_array_output = np.arange(n_axes_array_output)
    axes_other_output = axes_array_output[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_array_output, [axis_pooling_output, axis_receiving_output]))]
    shape_array_output = np.empty(n_axes_array_output, dtype=int)
    n_conditions_in_axes_merging_input = shape_array_input[axes_merging_input]
    shape_array_output[axis_pooling_output] = np.prod(n_conditions_in_axes_merging_input)
    shape_array_output[axis_receiving_output] = length_axis_receiving_output
    shape_array_output[axes_other_output] = shape_array_input[axes_other_array_input]
    if dtype is None:
        dtype = array_input.dtype
    array_output = np.empty(shape_array_output, dtype=dtype)

    indexes_in_axis_receiving_output = np.arange(length_axis_receiving_output)
    indexes_in_axis_receiving_output_from_axis_receiving_input = indexes_in_axis_receiving_output[np.logical_not(
        samples_in_arr1_are_in_arr2(indexes_in_axis_receiving_output, indexes_inserting_output))]

    # axis_variables = int(axis_receiving_input > axis_pooling_input)
    # axis_combinations = int(axis_variables == 0)
    axis_combinations = 0
    axis_variables = 1
    combinations_axes_removing_input = n_conditions_to_combinations(
        n_conditions_in_axes_removing_input, axis_combinations=axis_combinations, order_variables='lr')
    n_combinations_axes_removing_input = combinations_axes_removing_input.shape[axis_combinations]

    indexes_combinations = np.empty(2, dtype=object)

    indexes_input = np.empty(n_axes_array_input, dtype=object)
    indexes_input[:] = slice(0, None, 1)
    if length_axis_receiving_input == 1:
        indexes_input[axis_receiving_input] = 0

    indexes_output1 = np.empty(n_axes_array_output, dtype=object)
    indexes_output1[:] = slice(0, None, 1)
    indexes_output2 = np.copy(indexes_output1)
    indexes_output1[axis_receiving_output], indexes_output2[axis_receiving_output] = (
        seq_of_numeric_indexes_to_seq_of_slices(
            [indexes_in_axis_receiving_output_from_axis_receiving_input, indexes_inserting_output]))

    length_axis_pooling_input = shape_array_input[axis_pooling_input]
    start_index_axis_pooling_output = 0

    if n_axes_removing_input > 1:

        indexes_combinations[axis_variables] = slice(0, n_axes_removing_input, 1)

        shape_combinations_axes_removing_input_c = np.empty(n_axes_array_output, dtype=int)
        shape_combinations_axes_removing_input_c[:] = 1
        shape_combinations_axes_removing_input_c[axis_receiving_output] = n_indexes_inserting_output
        combinations_axes_removing_input_c = np.empty(shape_combinations_axes_removing_input_c, dtype=int)
        indexes_combinations_axes_removing_input_c = np.empty(n_axes_array_output, dtype=object)
        indexes_combinations_axes_removing_input_c[:] = 0
        indexes_combinations_axes_removing_input_c[axis_receiving_output] = slice(0, n_indexes_inserting_output, 1)
        indexes_combinations_axes_removing_input_c_tuple = tuple(indexes_combinations_axes_removing_input_c)

        for c in range(n_combinations_axes_removing_input):
            stop_index_axis_pooling_output = (c + 1) * length_axis_pooling_input
            indexes_output1[axis_pooling_output] = indexes_output2[axis_pooling_output] = slice(
                start_index_axis_pooling_output, stop_index_axis_pooling_output)
            start_index_axis_pooling_output = stop_index_axis_pooling_output

            indexes_combinations[axis_combinations] = c
            indexes_input[axes_removing_input] = (
                combinations_axes_removing_input[tuple(indexes_combinations)])
            array_output[tuple(indexes_output1)] = array_input[tuple(indexes_input)]

            combinations_axes_removing_input_c[indexes_combinations_axes_removing_input_c_tuple] = (
                combinations_axes_removing_input[tuple(indexes_combinations)])

            array_output[tuple(indexes_output2)] = combinations_axes_removing_input_c

    elif n_axes_removing_input == 1:

        indexes_combinations[axis_variables] = 0

        for c in range(n_combinations_axes_removing_input):
            stop_index_axis_pooling_output = (c + 1) * length_axis_pooling_input
            indexes_output1[axis_pooling_output] = indexes_output2[axis_pooling_output] = slice(
                start_index_axis_pooling_output, stop_index_axis_pooling_output)
            start_index_axis_pooling_output = stop_index_axis_pooling_output

            indexes_combinations[axis_combinations] = c
            indexes_input[axes_removing_input] = (
                combinations_axes_removing_input[tuple(indexes_combinations)])
            array_output[tuple(indexes_output1)] = array_input[tuple(indexes_input)]

            array_output[tuple(indexes_output2)] = combinations_axes_removing_input[tuple(indexes_combinations)]

    return array_output


def merge_neighbouring_conditions(array_input, axis_1, axis_2, m=3, s=None):
    # axis_1 is (are) the axis (axes) that contains (contain) the merging conditions. it can be
    #        an integer or a list of integers. In case of list of integers, the list cannot contain repeated values.
    # axis_2 is the axis that contains the merging samples of a set of n neighbouring conditions in axis_1.
    #        it is always an integer.
    # m is (are) the number(s) of the merging conditions in axis_1. it is an integer is axis_1 is an integer or
    #        a list of integers if axis_1 is a list of integers. In case of list of int, its size has to be the
    #        same as the size of the axis_1.

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = shape_array_input.size

    # format axes_2
    if axis_2 < 0:
        axis_2 += n_axes_array_input

    # format axes_1
    try:
        n_axis_1 = len(axis_1)
        axis_1 = np.asarray(axis_1, dtype=int)
        axis_1[axis_1 < 0] += n_axes_array_input
        # check point 1
        if np.sum(axis_1[0] == axis_1) > 1:
            raise ValueError('axis_1 cannot contain repeated values')
    except TypeError:
        if axis_1 < 0:
            axis_1 += n_axes_array_input
        axis_1 = np.asarray([axis_1], dtype=int)
        n_axis_1 = 1

    axes_merging = np.append(axis_2, axis_1)

    # format m
    try:
        n_m = len(m)
        if n_m == n_axis_1:
            m = np.asarray(m, dtype=int)
        elif n_m == 1:
            m_tmp = m[0]
            m = np.empty(n_axis_1, dtype=int)
            m[:] = m_tmp
            n_m = n_axis_1
        else:
            # check point 3
            raise ValueError('sizes of axis_1 and m must be equal')
    except TypeError:
        m_tmp = m
        m = np.empty(n_axis_1, dtype=int)
        m[:] = m_tmp
        n_m = n_axis_1

    # format s
    if s is None:
        s = m
    else:
        try:
            n_s = len(s)
            if n_s == n_axis_1:
                s = np.asarray(s, dtype=int)
                for a in range(n_axis_1):
                    if s[a] is None:
                        s[a] = m[a]
            elif n_s == 1:
                s_tmp = s[0]
                if s_tmp is None:
                    s = m
                else:
                    s = np.empty(n_axis_1, dtype=int)
                    s[:] = s_tmp
                n_s = n_axis_1
            else:
                # check point 3
                raise ValueError('sizes of axis_1 and s must be equal')
        except TypeError:
            s_tmp = s
            s = np.empty(n_axis_1, dtype=int)
            s[:] = s_tmp
            n_s = n_axis_1

    r_1 = m // 2
    r_0 = r_1
    m_even = (m % 2) == 0
    if any(m_even):
        r_0[m_even] -= 1

    n_conditions_input = shape_array_input[axis_1]
    n_samples_per_condition_input = shape_array_input[axis_2]

    n_conditions_output = ((n_conditions_input - m) // s) + 1
    n_samples_per_m_conditions_output = n_samples_per_condition_input * np.prod(m)

    shape_array_output = shape_array_input
    shape_array_output[axis_1] = n_conditions_output
    shape_array_output[axis_2] = n_samples_per_m_conditions_output
    array_output = np.empty(shape_array_output, dtype=array_input.dtype)

    index_input = np.empty(n_axes_array_input, dtype=object)
    index_input[:] = slice(None)
    index_output = np.copy(index_input)

    center = np.empty(n_axis_1, dtype=object)
    for a in range(n_axis_1):
        center[a] = np.arange(r_0[a], n_conditions_input[a] - r_1[a], s[a])
    combinations_centers = conditions_to_combinations(center)
    n_combinations_centers = combinations_centers.shape[0]

    combinations_merged_conditions_output = n_conditions_to_combinations(n_conditions_output)
    for i in range(n_combinations_centers):
        for a in range(n_axis_1):
            index_input[axis_1[a]] = slice(combinations_centers[i, a] - r_0[a], combinations_centers[i, a] + r_1[a] + 1)
        index_output[axis_1] = combinations_merged_conditions_output[i]

        # array_output[tuple(index_output)] = merge_axes(array_input[tuple(index_input)], axis_1, axis_2)

        array_output[tuple(index_output)] = merge_axes(array_input[tuple(index_input)], axes_merging)

    return array_output



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ccalafiore",
    version="000.000.056",
    author="Carmelo Calafiore",
    author_email="cc18849@essex.ac.uk",
    description="personal package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pypi.org/project/ccalafiore/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"],
    install_requires=[
        'numpy',
        'torch',
        'torchvision',
        'Pillow',
        'opencv-contrib-python',
        'matplotlib',
        'selenium'],
    python_requires='>=3.6')

# examples of the kw "install_requires"
# setup(
#     #...,
#     install_requires = [
#         'docutils',
#         'BazSpam ==1.1',
#         "enum34;python_version<'3.4'",
#         "pywin32 >= 1.0;platform_system=='Windows'"]
#     #...)

# more info the kw "install_requires" at the link below
# at https://setuptools.readthedocs.io/en/latest/userguide/dependency_management.html#declaring-dependencies
